/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rg3DRom.h"
#include "rg3DRom_csg.h"
#include "rgPort.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string.h>
#include "rg3DRom_file.h"


#define hlp_write_out_pending_tag sk3DRom_write_out_pending_tag
#define TEMP_WRITE_TAG_CODE( mv_des, mv_t)\
	rgWORD n = rgPortLittleEndianWORD( rgWORD( (mv_t).mode));\
	rgDWORD d = rgPortLittleEndianDWORD( rgDWORD( (mv_t).size));\
	(mv_des).write( (char *)&n, 2);\
	(mv_des).write( (char *)&d, 4);\
	if( (mv_des).fail()) return -1;

int sk3DRom_write_tag( std::ostream &des, sk3DRomTag &t)
{
	TEMP_WRITE_TAG_CODE( des, t);
	return 0;
}
#undef TEMP_WRITE_TAG_CODE

// Description: Writes the pending tag "t" to the file and upon returning,
//	leaves the file pointer at the same position it had when the function
//	was called. The pending tag "t" is written to whichever file position
//	is indiciated by its "size" member variable. Therefore, all tags
//	passed in must have the "size" member variable already set to its
//	file position. (The entire tag is written to the file, 6 bytes, at
//	the position indiciated by "size"). The tag's real size is written
//	to the file and the parameter "t" is untouched. 
//
// Return: On success, a non-negative number is returned indicating the size
//	of the tag. On error, a negative number is returned.
int sk3DRom_write_out_pending_tag( std::ostream &des, sk3DRomTag t)
{
	// Remember the current seek location.
	const int i = des.tellp();
	// Seek to the tag's write postion.
	des.seekp( t.size, std::ios::beg);
	// Set the tag's real size.
	t.size = (i - des.tellp()) - sk3DROMTAG_FSIZE;
	// Write the tag to the file.
	sk3DRom_write_tag( des, t);
	// Return to the original seek location.
	des.seekp( i, std::ios::beg);
	// Check for errors and return.
	if( des.fail()) return -1;
	return t.size;
}

int sk3DRomModel::save( const char *fname)
{
	if( !fname) return -1;

	// Perform pre-save processing (if pre-saving can be done).
	if( pre_save) {	
		const int err = pre_save( this);
		if( err < 0) return err;
	}

	// XXX: Open file here to make sure that it works. This is because it
	//	wastes the caller's time to do a lengthy file processing and
	//	then open the file after and discover that there is a file open
	//	error. It is better to discover the file open error now. (The
	//	output for this model will be saved at the end of this
	//	function).
	std::ofstream des_final( fname);
	if( !des_final.is_open()) return -2;

	// XXX: ofstream is extremly slow when random seeking and writing
	//	is done to a file. Therefore, the random seeking and writing is
	//	done to ostringstream instead, since it is most likely
	//	implemented in memory, or a medium that is much faster than
	//	what ofstream is implemented on. When all seeking and writing
	//	is done, the results are written to ofstream (des_final).
	std::ostringstream des( std::ostringstream::out |
		std::ostringstream::binary);

	//************ File Header ************//
	// Write out file identifier
	des.write( sk3DROM_HEADER, strlen( sk3DROM_HEADER));

	// Write out the mode tag.
	// Memorize the current file position and seek ahead (leaving enough
	// room for this tag) so that the tag can be written later when its
	// final size is determined.
	sk3DRomTag head_tag( sk3DROM_MODED, des.tellp());
	head_tag.save_space( des);

	// The header porition of the file has been written (seeked), now
	// coninuing to the data portion.
	//
	// NOTE: Form here on in, any tag can be written, and tags can be in
	//	any order (perferably in an order that allows for the file
	//	to be loaded/opened as quickly and efficiently as possible).

	//************ Data portion ************//
	
	// Write out vertex information.
	sk3DRomTag t( sk3DROM_VERTEX_3, des.tellp());
	t.save_space( des);
	if( write_vertex_data( des)) return -9;
	if( hlp_write_out_pending_tag( des, t) < 0) return -10;

	// Write out face information.
	t = sk3DRomTag( sk3DROM_FACE_2, des.tellp());
	t.save_space( des);
	if( write_face_data( des)) return -19;
	if( hlp_write_out_pending_tag( des, t) < 0) return -20;

	// Write out the skeleton data (main skeleton, not the frames,
	// animation, etc).
	t = sk3DRomTag( sk3DROM_MAIN_SKELETON, des.tellp());
	t.save_space( des);
	if( write_main_skeleton_data( des)) return -29;
	if( hlp_write_out_pending_tag( des, t) < 0) return -30;
	
	// Write out skeleton frame data.
	t = sk3DRomTag( sk3DROM_SKELFRAMES_2, des.tellp());
	t.save_space( des);
	if( write_skeleton_frame_data( des)) return -39;
	if( hlp_write_out_pending_tag( des, t) < 0) return -40;

	// Write out skeleton animation data.
	t = sk3DRomTag( sk3DROM_SKEL_ANIMATION, des.tellp());
	t.save_space( des);
	if( write_skeleton_animation_data( des)) return -39;
	if( hlp_write_out_pending_tag( des, t) < 0) return -40;

	// Write out texture data (if any).
	if( !(texture.empty())) {
		t = sk3DRomTag( sk3DROM_MODEL_TEXTURES, des.tellp());
		t.save_space( des);
		if( write_texture_data( des)) return -49;
		if( hlp_write_out_pending_tag( des, t) < 0) return -50;
	}

	// Write out group data (if any).
	if( !(group.empty())) {
		t = sk3DRomTag( sk3DROM_GROUPS_DATA, des.tellp());
		t.save_space( des);
		if( write_group_data( des)) return -59;
		if( hlp_write_out_pending_tag( des, t) < 0) return -60;
	}

	// Write out page data (if any).
	if( page_count() > 0) {
		t = sk3DRomTag( sk3DROM_PAGES_DATA, des.tellp());
		t.save_space( des);
		if( write_pages_data( des)) return -69;
		if( hlp_write_out_pending_tag( des, t) < 0) return -70;
	}

	// Write out page data (if any).
	if( m_csg->brush_count() > 0) {
		t = sk3DRomTag( sk3DROM_CSG_DATA, des.tellp());
		t.save_space( des);
		if( write_csg_data( des)) return -79;
		if( hlp_write_out_pending_tag( des, t) < 0) return -80;
	}

	// Write out custom data (if any).
	if( write_custom_data) {
		t = sk3DRomTag( sk3DROM_CUSTOM, des.tellp());
		t.save_space( des);

		// Write out the 64 byte identifier. sk3DRomModel always
		// writes out 64 null characters for the identifier.
		char blank[64];
		memset( blank, 0, 64);
		des.write( blank, 64);
		if( des.fail()) return -987;

		if( write_custom_data( des, this)) return -989;
		if( hlp_write_out_pending_tag( des, t) < 0) return -990;

	} else if( custom_tag_memory_storage) {
		t = sk3DRomTag( sk3DROM_CUSTOM, des.tellp());
		t.save_space( des);

		des.write( custom_tag_memory_storage,
			custom_tag_memory_length);
		if( des.fail()) return -1000;
		if( hlp_write_out_pending_tag( des, t) < 0) return -1010;
	}


	//************ The file has been written ************//
	if( hlp_write_out_pending_tag( des, head_tag) < 0) return -2000;

	des_final << des.str();
	des_final.close();

	return 0;
}

int sk3DRomModel::write_vertex_data( std::ostream &des) const
{
	write_dword( des, SK3DROM_VERTEX_VERSION_A);
	// Write out the total number of vertices in the model
	write_dword( des, rgDWORD( v.size()));

	// Write out all of the vertices
	const int istop = int( v.size());
	for( int i = 0; i < istop; i++) {
		(v[i]).write_binary( des);
		write_float( des, v[i].weight);
		write_dword( des, v[i].gid());
		write_dword( des, v[i].flags);
	}
	return 0;
}

int sk3DRomModel::write_face_data( std::ostream &des) const
{
	write_word( des, f.size());	// Write out the number of faces
	write_dword( des, SK3DROM_FACE_VERSION_B);

	// Write out all of the face data
	for( int i = 0; i < int( f.size()); i++) {
		//****** Writing out of a single face begins ******//
		// Write out the vertex id's
		write_word( des, f[i].vid[0]);
		write_word( des, f[i].vid[1]);
		write_word( des, f[i].vid[2]);

		// Write out 3 vertex normals
		(f[i].n[0]).write_binary( des);
		(f[i].n[1]).write_binary( des);
		(f[i].n[2]).write_binary( des);

		// Write out 3 vertex lights/colours (RGBA), 4 bytes per light
		des.write( (char *)(f[i].li), 12);
		if( des.fail()) return -30;

		write_word( des, f[i].tid);

		// Write out 3 mapping coordinates (8 bytes per map coord)
		{
			rgFLOAT temp;
			for( short j = 0; j < 6; j++) {
				temp = rgPortLittleEndianFLOAT(
					rgFLOAT( f[i].map[j]));
				des.write( (char *)&temp, 4);
			}
			if( des.fail()) return -50;
		}

		write_dword( des, f[i].flags);
		write_dword( des, f[i].gid);
		if( des.fail()) return -60;

		//****** Done, the face has been written. ******//
	}

	return 0;
}

int sk3DRomModel::write_main_skeleton_data( std::ostream &des) const
{
	// Write out the total number of bones.
	rgWORD n = rgPortLittleEndianWORD( rgWORD( skel.b.size()));
	des.write( (char *)&n, 2);
	if( des.fail()) return -10;

	// Write out all of the bones.
	// NOTE: All bones are preceeded by a sk3DROM_BONE tag.
	const int istop = int( skel.b.size());
	for( int i = 0; i < istop; i++) {
		sk3DRomTag t( sk3DROM_BONE_2, des.tellp());
		t.save_space( des);

		// Write out the bone version
		n = rgPortLittleEndianWORD( SK3DROM_BONE_VERSION_A);
		des.write( (char *)&n, sizeof( n));

		/////////////// BASIC BONE DATA BEGINS //////////////////
		// Write out the bones joints. The joints are relative to the
		// bones own joint coordinate system (jcs). The joints
		// coordinates relative to the standard coordinate system can
		// be computed through calculation of the full jcs (this is
		// to be done by the loader, as putting that information in
		// the file will just waste space).
		(skel.b[i].ja).write_binary( des);
		(skel.b[i].jb).write_binary( des);

		// Write out the bones joint coordinate system (jcs).
		(skel.b[i].jcs).write_binary( des);

		// Write out the pivot joint.
		des.write( &(skel.b[i].pj), 1);
		if( des.fail()) return -80;

		// Write out the index of the master bone.
		n = rgPortLittleEndianWORD( skel.b[i].mid);
		des.write( (char *)&n, 2);
		if( des.fail()) return -90;

		// Write out the number of slave bones.
		n = rgPortLittleEndianWORD( skel.b[i].link.size());
		des.write( (char *)&n, 2);
		if( des.fail()) return -100;
		// Write out the list of slave bones (if any).
		{
			// NOTE: For each slave bone, there is a 2 byte
			//	bone id and a byte joint id. That is, there
			//	are 3 bytes for every slave bone.
			const int jstop = int( skel.b[i].link.size());
			for( int j = 0; j < jstop; j++) {
				n = rgPortLittleEndianWORD(
					skel.b[i].link[j].bid);
				des.write( (char *)&n, 2);
				des.write( &(skel.b[i].link[j].jid), 1);
				if( des.fail()) return -105;
			}
		}

		// Write out the number of controlled verticies.
		n = rgPortLittleEndianWORD( skel.b[i].vid.size());
		des.write( (char *)&n, 2);
		if( des.fail()) return -110;
		// Write out the list of controlled verticies (if any).
		{
			const int jstop = int( skel.b[i].vid.size());
			for( int j = 0; j < jstop; j++) {
				n = rgPortLittleEndianWORD(
					skel.b[i].vid[j]);
				des.write( (char *)&n, 2);
				if( des.fail()) return -115;
			}
		}
		/////////////// BASIC BONE DATA ENDS //////////////////

		// Write out the name of the bone
		des.write( skel.b[i].bname, SK3DROM_BONE_NAME_LEN);

		// The data for this bone has been written, therefore write
		// out the tag for this bone.
		if( hlp_write_out_pending_tag( des, t) < 0) return -120;
	}

	return 0;
}

int sk3DRomModel::write_skeleton_frame_data( std::ostream &des) const
{
	// Write out the total number of skeleton frames.
	write_word( des, skel.sf.size());

	// Write out the total number of bones per skeleton.
	// XXX NOTE: There MUST be the same number of bones per frame as there
	//	are bones in the main skeleton. The total number of bones
	//	in each frame is written here again for redundancy and also
	//	to allow skeleton frames and the main skeleton tags to appear
	//	in the file in any order.
	write_word( des, skel.b.size());

	// Write out all of the skeleton frames now.
	// NOTE: Each skeleton frame is preceeded by an sk3DROM_SFRAME tag.
	const int istop = int( skel.sf.size());
	for( int i = 0; i < istop; i++) {
		
		sk3DRomTag tagFrame( sk3DROM_SFRAME, des.tellp());
		tagFrame.save_space( des);

		// Write out the name of this skeleton frame.
		des.write( skel.sf[i].fname, SK3DROM_SKEL_FRAME_NAME_LEN);
		if( des.fail()) return -30;

		// Write out all of the bones in this skeleton frame.
		// NOTE: Each bone in the skeleton frame is preceeded by an
		//	sk3DROM_BONE_FRAME_2 tag.
		const int jstop = int( skel.b.size());
		for( int j = 0; j < jstop; j++) {

			sk3DRomTag tagBone( sk3DROM_BONE_FRAME_2, des.tellp());
			tagBone.save_space( des);

			write_word( des, SK3DROM_BONE_FRAME_VERSION_A);

			const sk3DRomBoneFrame *b = &(skel.sf[i].b[j]);
			(b->rcs).write_binary( des);

			write_dword( des, b->msb);
			write_dword( des, b->mmb);
			write_dword( des, b->cms);
			if( des.fail()) return -60;

			//****** The bone was written ******//
			if( hlp_write_out_pending_tag( des, tagBone) < 0)
				return -75;
		}

		//************ The skeleton frame was written ************//
		if( hlp_write_out_pending_tag( des, tagFrame) < 0) return -100;
	}

	return 0;
}

int sk3DRomModel::write_skeleton_animation_data( std::ostream &des) const
{
	// Write out the total number of sequences.
	rgWORD n = rgPortLittleEndianWORD( skel.seq.size());
	des.write( (char *)&n, 2);
	if( des.fail()) return -10;

	// Each sequence (including its frames) will be encapsulated in an
	// sk3DROM_ANI_SEQ tag.
	const int istop = int( skel.seq.size());
	for( int i = 0; i < istop; i++) {

		sk3DRomTag t( sk3DROM_ANI_SEQ, des.tellp());
		t.save_space( des);

		// Write out the name of this sequence.
		des.write( skel.seq[i].get_name(), SK3DROM_SEQUENCE_NAME_LEN);
		if( des.fail()) return -30;

		// Write out all of the animation frames in this sequence.
		if( write_skeleton_animation_frame_data( des, i)) return -50;

		//************ The sequence was written ************//
		if( hlp_write_out_pending_tag( des, t) < 0) return -100;
	}

	return 0;
}

int sk3DRomModel::write_skeleton_animation_frame_data( std::ostream &des,
	const int sid) const
{
	// Write out the total number of animation frames.
	write_word( des, short( skel.seq[sid].frame.size()));
	for( int i = 0; i < int( skel.seq[sid].frame.size()); i++) {

		sk3DRomTag t( sk3DROM_ANI_FRAME_2, des.tellp());
		t.save_space( des);

		write_word( des, SK3DROM_ANI_FRAME_VERSION_A);
		write_dword( des, skel.seq[sid].frame[i].fid());
		write_word( des, skel.seq[sid].frame[i].time());
		write_word( des, skel.seq[sid].frame[i].code());
		write_dword( des, skel.seq[sid].frame[i].data[0]);
		write_dword( des, skel.seq[sid].frame[i].data[1]);
		write_dword( des, skel.seq[sid].frame[i].data[2]);
		write_dword( des, skel.seq[sid].frame[i].data[3]);
		write_dword( des, skel.seq[sid].frame[i].data[4]);
		write_dword( des, skel.seq[sid].frame[i].data[5]);

		//************ The sequence was written ************//
		if( hlp_write_out_pending_tag( des, t) < 0) return -100;
	}
	return 0;
}

int sk3DRomModel::write_texture_data( std::ostream &des) const
{
	// Write out the total number of textures.
	write_word( des, short( texture.size()));

	// Write out the textures now.
	const int istop = int( texture.size());
	for( int i = 0; i < istop; i++) {

		sk3DRomTag t( sk3DROM_TEXTURE_3, des.tellp());
		t.save_space( des);

		write_dword( des, SK3DROM_TEXTURE_VERSION_B);

		{ // Write out the textures name (there is no NULL terminator).
			const short len = short(
				texture[i].full_path_length());
			write_word( des, len);
			des.write( texture[i].full_path(), len);
		}

		{ // Write out the the name of the texture.
			const short len = short( texture[i].m_tname.length());
			write_word( des, len);
			des.write( texture[i].m_tname.c_str(), len);
		}

		write_dword( des, texture[i].flags());
		des.write( (char *)(texture[i]).m_colorkey, 4);

		//************ The texture was written ************//
		if( hlp_write_out_pending_tag( des, t) < 0) return -100;
	}
	return 0;
}

// NOTE: The below two macros are required because there is currently no
//	"known" way for converting between "std::fstream" and "std::ofstream".
//	Also, they are required for ensuring that there is only one copy of
//	the custom tag saving code.
#define WRITE_OUT_THE_TAG_HEADER( mv_des, mv_t) \
	if( (mv_t).size < 0) return -3; \
	if( sk3DRom_write_tag( (mv_des), (mv_t))) return -4;

#define WRITE_OUT_THE_TAGS_REMAINING_DATA( mv_des, mv_src, mv_t) \
	rgDWORD data_remain = rgDWORD( (mv_t).size) + \
		sk3DROMTAG_FSIZE;\
	const rgDWORD buf_size = 16384;\
	char *buf = new char[ buf_size];\
	while( data_remain) {\
		if( data_remain > buf_size) {\
			(mv_src).read( buf, buf_size);\
			if( (mv_src).fail()) return -10;\
			(mv_des).write( buf, buf_size);\
			if( (mv_des).fail()) return -11;\
			data_remain -= buf_size;\
		} else {\
			(mv_src).read( buf, data_remain);\
			if( (mv_src).fail()) return -20;\
			(mv_des).write( buf, data_remain);\
			if( (mv_des).fail()) return -21;\
			data_remain = 0;\
		}\
	}\
	delete [] buf;


int sk3DRomModel::store_custom_tag( std::iostream &des, std::ifstream &src,
	sk3DRomTag &t)
{
	WRITE_OUT_THE_TAG_HEADER( des, t);
	WRITE_OUT_THE_TAGS_REMAINING_DATA( des, src, t);
	return 0;
}

/*
int sk3DRomModel::store_custom_tag( std::ofstream &des, std::ifstream &src,
	sk3DRomTag &t)
{
	WRITE_OUT_THE_TAG_HEADER( des, t);
	WRITE_OUT_THE_TAGS_REMAINING_DATA( des, src, t);
	return 0;
}
*/

int sk3DRomModel::write_group_data( std::ostream &des) const
{
	// Write out the total number of goups.
        rgDWORD d = rgPortLittleEndianDWORD( group.size());
        des.write( (char *)&d, sizeof( d));
        if( des.fail()) return -10;

	// Write out the groups now.
	const int istop = int( group.size());
	for( int i = 0; i < istop; i++) {

		sk3DRomTag t( sk3DROM_GROUP, des.tellp());
		t.save_space( des);

		// Write out the group name and colour.
		// NOTE: Less tests are done to make the code easier to read
		//	(without reducing security).
		des.write( group[i].gname, SK3DROM_GROUP_NAME_MAX_LEN);
		des.write( (char *)(group[i].gcolor), 4);
		d = rgPortLittleEndianDWORD( group[i].flags);
		des.write( (char *)&d, sizeof( d));
		if( des.fail()) return -15;
		// List of owned vertices are not written out (to save space).
		// Also, this information can be determined from the vertices
		// themselves anyway.

		//************ The group was written ************//
		if( hlp_write_out_pending_tag( des, t) < 0) return -100;
	}

	return 0;
}

int sk3DRomModel::write_pages_data( std::ostream &des) const
{
	write_dword( des, page_count());
	for( int i = 0; i < page_count(); i++) {
		sk3DRomTag t( sk3DROM_PAGE, des.tellp());
		t.save_space( des);

		write_dword( des, SK3DROM_PAGE_VERSION_A);
		des.write( m_page[i].page_name(), SK3DROM_TEX_PAGE_NAME_LEN);
		write_dword( des, m_page[i].flags());
		write_dword( des, m_page[i].tex_unit());
		write_float( des, m_page[i].uscale());
		write_float( des, m_page[i].vscale());
		write_dword( des, m_page[i].texid());
		write_page_faces( des, m_page[i]);
		if( des.fail()) return -15;

		//************ The page was written ************//
		if( sk3DRom_write_out_pending_tag( des, t) < 0) return -100;
	}

	return 0;
}

void sk3DRomModel::write_page_faces( std::ostream &des,
	const sk3DRomTexturePage &p) const
{
	write_dword( des, p.face_count());
	for( int i = 0; i < p.face_count(); i++) {
		const sk3DRomTextureFace *tmp = p.face( i);
		write_dword( des, tmp->fid());
		(tmp->v( 0)).write_binary( des);
		(tmp->v( 1)).write_binary( des);
		(tmp->v( 2)).write_binary( des);
	}
}

int sk3DRomModel::write_csg_data( std::ostream &des) const
{
	write_dword( des, m_csg->brush_count());
	for( int i = 0; i < m_csg->brush_count(); i++) {
		sk3DRomTag t( sk3DROM_CSG_BRUSH, des.tellp());
		t.save_space( des);

		write_dword( des, (m_csg->brush( i))->type());
		(m_csg->brush( i))->save_data( des);

		//************ The brush was written ************//
		if( sk3DRom_write_out_pending_tag( des, t) < 0) return -100;
	}
	return 0;
}





