/*
Copyright (C) 1997-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream>
#include <fstream>
#include <string.h>
#include "rgError.h"
#include "rgBMP.h"

/***************************************************************************
 * Little endian definitions begin
 ***************************************************************************/
#ifdef RGPORT_LITTLE_ENDIAN
#define READ_WORD(n) bmp.read( (char *)&(n), 2)
#define READ_DWORD(n) bmp.read( (char *)&(n), 4)
#endif
/***************************************************************************
 * Little endian definitions end
 ***************************************************************************/

/***************************************************************************
 * Big endian definitions begin
 ***************************************************************************/
#ifdef RGPORT_BIG_ENDIAN
#define READ_WORD(n) {bmp.read( (char *)&(n), 2); \
	(n) = rgPortLittleToBigWORD( n);}

#define READ_DWORD(n) {bmp.read( (char *)&(n), 4); \
	(n) = rgPortLittleToBigDWORD( n);}
#endif
/***************************************************************************
 * Big endian definitions end
 ***************************************************************************/

rgBumpMap::rgBumpMap()
{
	width = 0;
	height = 0;
	data = 0;
	nmap = 0;
}

rgBumpMap::~rgBumpMap()
{
	clear();
}
                                                                                
void rgBumpMap::clear()
{
	if( data) { delete [] data; data = 0;}
	if( nmap) { delete [] nmap; nmap = 0;}
}
                                                                                
int rgBumpMap::load( const char *fname)
{
	clear();

	if( !fname) return RGERR_BAD_PARAM;

	rgBMP bmp;
	int err = bmp.load( fname);
	if( err != RGERR_SUCCESS) return err;

	/* Bump map image must be 8 bit. */
	if( bmp.getColorDepth() != 8) return RGERR_FAIL;

	width = bmp.getWidth();
	height = bmp.getHeight();
	const int data_size = width * height;
	data = new rgFloat[ data_size];
	rgBYTE *p = bmp.getImageData();
	int i;
	const rgFloat k = 1.0f/255.0f;
	for( i = 0; i < data_size; i++)
		data[i] = rgFloat( p[i]) * k;

	return RGERR_SUCCESS;
}

void rgBumpMap::scaleData( const rgFloat k)
{
	if( !data) return;
	int i = width * height;
	while( i)
		data[ --i] *= k;
}

void rgBumpMap::scaleNormals( const rgFloat x, const rgFloat y, const rgFloat z)
{
	if( !nmap) return;
	int i = width * height;
	while( i) {
		i--;
		nmap[i].x = nmap[i].x * x;
		nmap[i].y = nmap[i].y * y;
		nmap[i].z = nmap[i].z * z;
	}
}

void rgBumpMap::normalize()
{
	if( !nmap) return;
	int i = width * height;
	while( i) nmap[ --i].normalize();
}

void rgBumpMap::invert()
{
	if( !data) return;
	int i = width * height;
	while( i) {
		data[i] = 1.0f - data[i];
		i--;
	}
}

void rgBumpMap::genNormalMap()
{
	if( !data) return;

	if( nmap) delete nmap;
	nmap = new rgVector[ width * height];

	rgVector n;
	rgVector bn( 0.0f, 0.0f, -1.0f);
	rgFloat *dptr = data;
	rgVector *nptr = nmap;
	int x, y;

	/* Generate normals by only considering the horizontal (X axis)
	 * relationship between the bump map sample points.
	 */
	//#ifdef dfdg
	for( y = 0; y < height; y++) {
	
		/* Generate the first normal in the row. */	
		n = bn.cross( 1.0f, dptr[1] - dptr[0], 0.0f).normalize();
		nptr[0] = n;

		/* Generate row normals (except first and last). */
		for( x = 1; x < (width - 1); x++) {
			nptr[x] = n;
			n = bn.cross( 1.0f, dptr[ x + 1] - dptr[x],
				0.0f).normalize();
			nptr[x] = (nptr[x] + n) * 0.5f;
		}

		/* Generate the last normal in the row. */
		nptr[x] = n;

		nptr += width;
		dptr += width;
	}
	//#endif

	/* Finalize normals by considering the vertical (Y axis) relationship
	 * between the bump map sample points.
	 */
	//#ifdef dfads
	const int ytemp = width * (height - 1);
	int ystop;
	bn = rgVector( -1.0f, 0.0f, 0.0f);
	dptr = data;
	nptr = nmap;
	rgVector n2;
	for( x = 0; x < width; x++) {
	
		/* Generate the first normal in the column. */
		n = bn.cross( 0.0f, dptr[width] - dptr[0],
			-1.0f).normalize();
		nptr[0] = (nptr[0] + n) * 0.5f;

		/* Generate column normals (except first and last). */
		ystop = ytemp;
		for( y = width; y < ystop; y += width) {
			n2 = nptr[y];
			nptr[y] = n;
			n = bn.cross( 0.0f, dptr[ y + width] - dptr[y],
				-1.0f).normalize();
			nptr[y] = (nptr[y] + n + n2) * 0.33333f;
		}

		/* Generate the last normal in the column. */
		nptr[y] = (nptr[y] + n) * 0.5f;

		nptr++;
		dptr++;
	}
	//#endif

        // The above code generates normals that point in the wrong
        // direction. Inorder to make rgBumpMap less confusing to use, the
        // normals should point in the correct direction (that is, point
        // up towards positive Y instead of down towards negative Y).
        scaleNormals( -1, -1, -1);
}


#define NORMAL_COMPONENT_TO_COLOR(n) \
	RG_MIN( 255, RG_MAX( 0, (rgBYTE)(((n) + 1.0f) * 127.5f)))
int rgBumpMap::gen_texture_bump_map( rgBMP &bmp) const
{
	if( (width < 0) || (height < 0)) return -1;
	if( !nmap) return -2;
	if( bmp.generate_bmp( width, height, 24)) return -3;
	rgBYTE *row = bmp.getImageData();
	if( !row) return -4;
	const rgVector *n = nmap;
	for( int y = 0; y < height; y++) {
		for( int x = 0; x < width; x++) {
			*(row++) = NORMAL_COMPONENT_TO_COLOR( n[x].x);
			*(row++) = NORMAL_COMPONENT_TO_COLOR( n[x].z);
			*(row++) = NORMAL_COMPONENT_TO_COLOR( n[x].y);
		}
		n += width;
	}
	return 0;
}
#undef NORMAL_COMPONENT_TO_COLOR


rgBMP::rgBMP()
{
	width = 0;
	height = 0;
	palette = 0;
	imageData = 0;
}

rgBMP::~rgBMP()
{
	clear();
}

void rgBMP::clear()
{
	width = 0;
	height = 0;
	colorDepth = 0;

	if( palette) { delete [] palette; palette = 0;}
	if( imageData) { delete [] imageData; imageData = 0;}
}

int rgBMP::query_width_and_height( const char *fname, int &p_width,
	int &p_height)
{
	if( !fname) return RGERR_BAD_PARAM;

	std::ifstream bmp( fname, std::ios::binary);
	if( !bmp.is_open()) return RGERR_COULD_NOT_OPEN_FILE;

	/* Check the file's type (file should begin with "BM" as the first
	 * 2 characters).
	 */
	char bmpID[2];
	bmp.read( bmpID, sizeof( bmpID));
	if( strncmp( bmpID, "BM", sizeof( bmpID)) != 0) return -1001;

	bmp.seekg( BMP_WIDTH_OFF, std::ios::beg);
	READ_DWORD( p_width);

	bmp.seekg( BMP_HEIGHT_OFF, std::ios::beg);
	READ_DWORD( p_height);

	return RGERR_SUCCESS;
}

int rgBMP::load( const char *fname)
{
	return load( fname, 0, 0, 0);
}

int rgBMP::load( const char *fname, int flags)
{
	return load( fname, flags, 0, 0);
}

int rgBMP::load( const char *fname, int flags, int maxWidth, int maxHeight)
{
	if( !fname) return RGERR_BAD_PARAM;

	/*******************************************************************
	 * Check the flags given.
	 *******************************************************************/
	/* If an RGBA bitmap was requested, then this implies that RGB was
	 * requested as well. This may seem strange, but it is because this
	 * function creates RGBA bitmaps from RGB bitmaps. This allows for
	 * easier coding since the function can convert any bitmap type to
	 * RGB and from there it can create an RGBA bitmap. That is, it allows
	 * for code reuse.
	 */
	if( (flags & RGBMP_FORCE_RGBA)) flags |= RGBMP_FORCE_RGB;
	
	/* Delete current BMP information. */
	clear();

	/********************************************************************
	 * Open the bmp file.
	 *
	 * NOTE: Any opened file is automatically closed when the "bmp" class's
	 *	destructor is called. That is, when this function returns, it
	 *	automatically closes any opened file. So you will not see a
	 *	bmp.close() statement on every return statment. Near the end
	 *	of the function, bmp.close() is called only because the file
	 *	is no longer needed. That is, the function is being "nice" and
	 *	gives up its lock on the file so others may access it. Not
	 *	putting bmp.close() on every return makes the code shorter and
	 *	easier to read.
	 ********************************************************************/
        std::ifstream bmp( fname, std::ios::binary);
	if( !bmp.is_open()) return RGERR_COULD_NOT_OPEN_FILE;

	/* Check the file's type (file should begin with "BM" as the first
	 * 2 characters).
	 */
	char bmpID[2];
	bmp.read( bmpID, sizeof(bmpID));
	if( strncmp( bmpID, "BM", sizeof(bmpID)) != 0) return -1001;

        bmp.seekg( BMP_WIDTH_OFF, std::ios::beg);
	READ_DWORD( width);

        bmp.seekg( BMP_HEIGHT_OFF, std::ios::beg);
	READ_DWORD( height);

	if( width <= 0) return -1002;
	if( height <= 0) return -1003;

	if( maxWidth) { if( width > maxWidth) return -1004;}
	if( maxHeight) { if( height > maxHeight) return -1005;}

        bmp.seekg( BMP_BITCOUNT_OFF, std::ios::beg);
	READ_WORD( colorDepth);

	/* Allocate memory to store the image data */
	int imageDataSize;
	switch( colorDepth) {
	case 8:
		/* 8-bit bmp's have a 256 colour palette */
		palette = new rgBMPPaletteEntry[256];
		if( !palette) return -1006;

		/* Read 256 color entiries from the palette. Each entry is
		 * 4 bytes long, so a total of 1024 bytes will be read.
		 */
                bmp.seekg( BMP_PALETTE_OFF, std::ios::beg);
		bmp.read( (char *)palette, 1024);

		if( (flags & RGBMP_FORCE_RGB)) {
			if( (flags & RGBMP_FORCE_RGBA)) {
				imageDataSize = width * height * 4;
			} else imageDataSize = width * height * 3;
		} else imageDataSize = width * height;
		break;
	case 16:
		if( (flags & RGBMP_FORCE_RGB)) {
			if( (flags & RGBMP_FORCE_RGBA)) {
				imageDataSize = width * height * 4;
			} else imageDataSize = width * height * 3;
		} else imageDataSize = width * height * 2;
		break;
	case 24:
		if( (flags & RGBMP_FORCE_RGBA)) {
			imageDataSize = width * height * 4;
		} else imageDataSize = width * height * 3;
		break;
	default: return -1007;
	}

	imageData = new rgBYTE[ imageDataSize];
	if( !imageData) {
		if( palette) { delete [] palette; palette = 0;}
		return -1008;
	}

	/********************************************************************
	 * Read in the image data
	 ********************************************************************/
	int dataOffset;
        bmp.seekg( BMP_DATABEGIN_OFF, std::ios::beg);
	READ_DWORD( dataOffset);
        bmp.seekg( dataOffset, std::ios::beg);

	int error;
	switch( colorDepth) {
	case 8:	/* Read in 8 bit image data */
		if( (flags & RGBMP_FORCE_RGB)) error = read8to24( bmp);
		else error = read8to8( bmp);
		break;
	case 16: /* Read in 16 bit image data */
		if( (flags & RGBMP_FORCE_RGB)) error = read16to24( bmp);
		else error = read16to16( bmp);
		break;
	case 24: /* Read in 24 bit image data */
		error = read24to24( bmp);
		break;
	default:
		/* Previous colorDepth check will prevent execution from
		 * ever reaching this point. If execution does reach this
		 * point, then there is a serious programmer error.
		 */
		error = -1009;
		break;
	}

	/* Finished reading file, therefore close it to allow others to
	 * access it again.
	 */
	bmp.close();

	/* Check if the image data was loaded. */
	if( error != 0) {
		if( palette) { delete [] palette; palette = 0;}
		if( imageData) { delete [] imageData; imageData = 0;}

		return error;
	}

	/* Produce 32-bit RGBA output (if it was requested).
	 * NOTE: If RGBA output was requested then the image data is curently
	 * 	in RGB format, and the image data array was allocated to
	 * 	provide enough space for RGBA expansion.
	 */
	if( (flags & RGBMP_FORCE_RGBA)) {
		error = convert24to32( width, height, imageData, imageData);
		if( error != RGERR_SUCCESS) {
			if( palette) { delete [] palette; palette = 0;}
			if( imageData) { delete [] imageData; imageData = 0;}
			
			return -1011;
		}
	}
	
	return RGERR_SUCCESS;
}

void rgBMP::get_sub_image( rgBYTE *p_imgData, int x1, int y1, int x2,
	int y2, int row_stride)
{
	/* Get the size of the pixel data. */
	const int spd = colorDepth >> 3;
	if( spd <= 0) return;

	const int src_width = spd * width;
	rgBYTE *img = imageData + (y1 * src_width);
	rgBYTE *src;

	/* FIXME: Hack. Why do we have to increment row_stride? */
	if( row_stride) row_stride++;	/* ??? */
	const int rs = spd * row_stride;

	int x, y;
	
	if( x2 > (width - 1)) x2 = width - 1;
	if( y2 > (height - 1)) y2 = height - 1;
	for( y = y1; y < y2; y++) {

		src = img + (x1 * spd);
		for( x = x1; x < x2; x++) {
			memcpy( p_imgData, src, spd);

			src += spd;
			p_imgData += spd;
		}

		img += src_width;
		p_imgData += rs;
	}
}

void rgBMP::flip_vertically()
{
	if( !imageData) return;
	if( height <= 0) return;

	// Swap rows
	int step;
	switch( colorDepth) {
	case 8: step = width; break;
	case 16: step = width << 1; break;
	case 24: step = width * 3; break;
	default: return;
	}

	rgBYTE *rowa = imageData;
	rgBYTE *rowb = imageData + ((height - 1) * step);
	rgBYTE temp;
	int i;
	while( rowa < rowb) {

		// Swap the rows;
		for( i = 0; i < step; i++) {
			temp = rowa[i];
			rowa[i] = rowb[i];
			rowb[i] = temp;
		}
		
		rowa += step;
		rowb -= step;
	}
}

int rgBMP::read8to8( std::ifstream &bmp)
{
	int y;
	rgBYTE *p;

	/* Allocate memory for an 8-bit image. */
	p = imageData + ((height - 1) * width);
	for( y = 0; y < height; y++) {
		/* Read an entire row and skip the padding bytes (if any). */
		bmp.read( (char *)p, width);
                bmp.seekg( (width % 4), std::ios::cur);
		
		p -= width;
	}

	return 0;
}

int rgBMP::read16to16( std::ifstream &bmp)
{
	int y;
	rgBYTE *p;

	/* Allocate memory for a 16-bit image. */	
	int w = width*2;
	p = imageData + ((height - 1) * w);
	for( y = 0; y < height; y++) {
		/* Read an entire row and skip the padding bytes (if any). */
		bmp.read( (char *)p, w);
                bmp.seekg( (w % 4), std::ios::cur);
		
		p -= w;
	}

	#ifdef RGPORT_BIG_ENDIAN
	/* BMP data is always in little endian. 16-bit data must be converted
	 * to big-endian for machines that require it, otherwise programs will
	 * interpret the 16-bit colour incorrectly. This is because 16 bit
	 * data is saved as a word rather than 2 bytes. 24-bit data does not
	 * have this problem since it is saved as 3 bytes, rather than a
	 * 32-bit double word (which would suffer from endian problems).
	 */
	for( y = 0; y < width*height; y++) {
		rgPortLittleToBigWORD( imageData[ y]);
	}
	#endif

	return 0;
}

int rgBMP::read24to24( std::ifstream &bmp)
{
	int y;
	rgBYTE *p;
	
	int w = width*3;
	const int padding = (4 - (w % 4)) & 0x3;

	p = imageData + ((height - 1) * w);
	for( y = 0; y < height; y++) {
		/* Read an entire row and skip the padding bytes (if any). */
		bmp.read( (char *)p, w);
                bmp.seekg( padding, std::ios::cur);
		
		p -= w;
	}

	return 0;
}

int rgBMP::read8to24( std::ifstream &bmp)
{
	rgBYTE *buffer = new rgBYTE[width];
	if( !buffer) return -1;

	int y, x;
	rgBYTE *p, *buf;

	const int padding = (4 - (width % 4)) & 0x3;

	int w = width*3;
	p = imageData + ((height - 1) * w);
	for( y = 0; y < height; y++) {
		/* Read an entire row and skip the padding bytes (if any). */
		bmp.read( (char *)buffer, width);
                bmp.seekg( padding, std::ios::cur);

		/* Convert the 8-bit row into 24-bit row */
		buf = buffer;
		for( x = 0; x < w; x += 3) {
			p[x] = palette[ *(buf)].red;
			p[x+1] = palette[ *(buf)].green;
			p[x+2] = palette[ *(buf)].blue;

			/* Move on to next 8-bit pixel */
			buf++;
		}
		
		p -= w;	/* Move on to previous row */
	}

	colorDepth = 24;

	return 0;
}

int rgBMP::read16to24( std::ifstream &bmp)
{
	rgWORD *buffer = new rgWORD[width];
	if( !buffer) return -1;

	int y, x;
	rgBYTE *p;
	rgWORD *buf;

	int w = width*3;
	const int padding = (4 - ((width*2) % 4)) & 0x3;

	p = imageData + ((height - 1) * w);
	for( y = 0; y < height; y++) {
		/* Read an entire row and skip the padding bytes (if any). */
		bmp.read( (char *)buffer, width*2);
                //bmp.seekg( ((width*2) % 4), std::ios::cur);
                bmp.seekg( padding, std::ios::cur);
		
		/* Convert the 16-bit row into a 24-bit row. */
		buf = buffer;
		for( x = 0; x < w; x += 3) {

			/* NOTE: BMP data is always in little endian. */
			#ifdef RGPORT_BIG_ENDIAN
			rgPortLittleToBigWORD( buf[0]);
			#endif

			/* 16-bit BMP data is in the format 565. That is,
			 * 5 bits red, 6 bits green, and 5 bits blue.
			 */
			p[x] = ((buf[0] & 0xF800)>>11) << 3;
			p[x+1] = ((buf[0] & 0x07E0)>>5) << 2;
			p[x+2] = (buf[0] & 0x001F) << 3;

			/* Move on to the next 16-bit pixel data. */
			buf++;
		}

		p -= w;	/* Move on to previous row */
	}

	colorDepth = 24;

	return 0;
}

void rgBMP::convertBGRtoRGB()
{
	convertBGRtoRGB( imageData, width * height);
}

void rgBMP::convertBGRtoRGB( rgBYTE *p_imgData,
	const int p_pixelCount)
{
	switch( colorDepth) {
	case 8:
		/* FIXME: Should the palette be swaped? */
		break;

	case 16:
		/* FIXME: Add support for 16 bit images. */
		break;

	case 24:
		{
			rgBYTE temp;
			int i;
			for( i = 0; i < p_pixelCount; i++) {
				temp = *(p_imgData);
				*(p_imgData) = *(p_imgData + 2);
				*(p_imgData + 2) = temp;
				p_imgData += 3;
			}
		}
		break;

	case 32:
		{
			rgBYTE temp;
			int i;
			for( i = 0; i < p_pixelCount; i++) {
				temp = *(p_imgData);
				*(p_imgData) = *(p_imgData + 2);
				*(p_imgData + 2) = temp;
				p_imgData += 4;
			}
		}
		break;

	}
}

int rgBMP::convert16to24( const int p_width, const int p_height,
	rgBYTE *img24, rgWORD *img16, int p_format)
{
	if( p_width <= 0) return RGERR_BAD_PARAM;
	if( p_height <= 0) return RGERR_BAD_PARAM;
	if( !img24) return RGERR_BAD_PARAM;
	if( !img16) return RGERR_BAD_PARAM;

	/* Point to the begining of the last RGB block (that is, point to
	 * the red component of the last RGB block).
	 */
	rgWORD *rgb16Ptr = img16 + ((p_width - 1) * p_height);

	/* Point to the begining of the last RGBA block (that is, point to
	 * the red component of the last RGBA block).
	 */
	rgBYTE *rgb24Ptr = img24 + ((p_width - 1) * p_height * 3);

	/* NOTE: This functions supports converting 16 bit to 24 bit images
	 *	even if img16 and img24 point to the same array (memory
	 *	location). The below comment explains how this is possible.
	 */
	/* Begin the RGB16 to RGB24 (RGB) conversion. This is done by working
	 * backwards (since the back part of the array is unused). The last
	 * RGB16 block is written to the last RGB24 block.
	 * Then the program moves on to the next RGB16 and RGB24
	 * block. When the function gets to the first RGB16 block the whole
	 * array will have been converted to a RGB24.
	 */

	switch( p_format) {
	case RGBMP_16BIT_1555:
		while( rgb16Ptr >= img16) {

			rgb24Ptr[0] = (rgb16Ptr[0] >> 10) & 0x1F;
			rgb24Ptr[1] = (rgb16Ptr[0] >> 5) & 0x1F;
			rgb24Ptr[2] = rgb16Ptr[0] & 0x1F;

			rgb16Ptr--;
			rgb24Ptr -= 3;
		}
		break;

	case RGBMP_16BIT_5551:
		while( rgb16Ptr >= img16) {

			rgb24Ptr[0] = (rgb16Ptr[0] >> 11) & 0x1F;
			rgb24Ptr[1] = (rgb16Ptr[0] >> 6) & 0x1F;
			rgb24Ptr[2] = (rgb16Ptr[0] >> 1) & 0x1F;

			rgb16Ptr--;
			rgb24Ptr -= 3;
		}
		break;

	case RGBMP_16BIT_565:
		while( rgb16Ptr >= img16) {

			rgb24Ptr[0] = (rgb16Ptr[0] >> 11) & 0x1F;
			rgb24Ptr[1] = (rgb16Ptr[0] >> 5) & 0x3F;
			rgb24Ptr[2] = rgb16Ptr[0] & 0x1F;

			rgb16Ptr--;
			rgb24Ptr -= 3;
		}
		break;

	default:
		return RGERR_BAD_PARAM;
	}

	colorDepth = 24;

	return RGERR_SUCCESS;
}

int rgBMP::convert24to32( const int p_width, const int p_height,
	rgBYTE *img32, rgBYTE *img24)
{
	if( p_width <= 0) return RGERR_BAD_PARAM;
	if( p_height <= 0) return RGERR_BAD_PARAM;
	if( !img32) return RGERR_BAD_PARAM;
	if( !img24) return RGERR_BAD_PARAM;

	/* Point to the begining of the last RGB block (that is, point to
	 * the red component of the last RGB block).
	 */
	rgBYTE *rgbPtr = img24 + (p_height * (p_width * 3) - 3);

	/* Point to the begining of the last RGBA block (that is, point to
	 * the red component of the last RGBA block).
	 */
	rgBYTE *rgbaPtr = img32 + (p_height * (p_width * 4) - 4);

	/* NOTE: This functions supports converting 24 bit to 32 bit images
	 *	even if img24 and img32 point to the same array (memory
	 *	location). The below comment explains how this is possible.
	 */
	/* Begin the RGB to RGBA conversion. This is done by working backwards
	 * (since the back part of the array is unused). The last RGB block
	 * is written to the last RGBA block. The alpha component is set to
	 * 0. Then the program moves on to the next RGB and RGBA block. When
	 * the function gets to the first RGB block the whole array will have
	 * been converted to a RGBA.
	 */
	while( rgbPtr >= img24) {
		rgbaPtr[0] = rgbPtr[0];
		rgbaPtr[1] = rgbPtr[1];
		rgbaPtr[2] = rgbPtr[2];
		rgbaPtr[3] = 0;

		rgbPtr -= 3;
		rgbaPtr -= 4;
	}

	colorDepth = 32;

	return RGERR_SUCCESS;
}

#ifdef RGPORT_BIG_ENDIAN
#define SET_AND_WRITE_DATA( data, value) \
	{ \
		(data) = (value); \
		(data) = rgPortLittleToBig( data); \
		bmp.write( (char *)&data, sizeof( data)); \
	}
#endif

#ifdef RGPORT_LITTLE_ENDIAN
#define SET_AND_WRITE_DATA( data, value) \
	{ \
		(data) = (value); \
		bmp.write( (char *)&data, sizeof( data)); \
	}

#endif

int rgBMP::save( const char *fname, const int p_width, const int p_height,
	const int p_colour_depth, const rgBYTE *p_bmpData,
	const rgBYTE *p_palette) const
{
	if( !fname) return RGERR_BAD_PARAM;
	if( p_width <= 0) return RGERR_BAD_PARAM;
	if( p_height <= 0) return RGERR_BAD_PARAM;

	switch( p_colour_depth) {
	case 1:
	case 2:
	case 4:
	case 8:
		if( !p_palette) return RGERR_BAD_PARAM;
		break;
	}

	if( (p_width > 0) && (p_height > 0)) {
		if( !p_bmpData) return RGERR_BAD_PARAM;
	}

	/********************************************************************
	 * Open the specified bitmap file.
	 ********************************************************************/
	std::ofstream bmp( fname, std::ios::binary);
	if( !bmp.is_open()) return RGERR_COULD_NOT_OPEN_FILE;

	rgWORD wordData;
	rgDWORD dwordData;

	/*******************************************************************
	 * Write out the bitmap info header.
	 *******************************************************************/
	bmp.seekp( 14, std::ios::beg);

	/* specifies the size of the BITMAPINFOHEADER structure in bytes.
	 * Standard valueis 40.
	 */
	SET_AND_WRITE_DATA( dwordData, 40);

	/* Write out the image width. */
	SET_AND_WRITE_DATA( dwordData, p_width);

	/* Write out the image height. */
	SET_AND_WRITE_DATA( dwordData, p_height);

	/* Write out number of planes. Standard value is 1. */
	SET_AND_WRITE_DATA( wordData, 1);

	/* Write out the colour depth (bits per pixel). */
	SET_AND_WRITE_DATA( wordData, p_colour_depth);

	/* Write out the compression type. Currently no compression is done. */
	SET_AND_WRITE_DATA( dwordData, 0);

	/* Write out the size of the image. */
	SET_AND_WRITE_DATA( dwordData, p_width * p_height);

	/* Write out the X pels per meter. Currently set to 0. */
	SET_AND_WRITE_DATA( dwordData, 0);
 
	/* Write out the Y pels per meter. Currently set to 0. */
	SET_AND_WRITE_DATA( dwordData, 0);

	/* Write out the number of colours used. Currently, 0 is written out
	 * indicating that the bmp loader must calculate the total number of
	 * colours used by checking the number of bits per pixel for the
	 * image.
	 */
	SET_AND_WRITE_DATA( dwordData, 0);

	/* Write out the number of important colours. Currently, 0 is written
	 * out, indicating that all colours are important.
	 */
	SET_AND_WRITE_DATA( dwordData, 0);

	/* Size of the bitmap file after the info header has been written. */
	int fileSize = 52;


	/*******************************************************************
	 * Write out the bitmap image data.
	 *******************************************************************/
	int paddingLength;
	int imageDataOffset = 0;
	int y;
	rgBYTE padding[3];

	const rgBYTE *ptr;
	switch( p_colour_depth) {
	case 24:
		imageDataOffset = bmp.tellp();
		paddingLength = (4 - ((p_width * 3) & 0x3)) & 0x3;
		for( y = p_height - 1; y >= 0; y--) {

			ptr = p_bmpData + (y * p_width * 3);
			{
				char temp[3];
				for( int i = 0; i < p_width; i++) {
					// Swap RGB to BGR.
					temp[0] = ptr[2];
					temp[1] = ptr[1];
					temp[2] = ptr[0];
					bmp.write( temp, 3);
					ptr += 3;
				}
			}
			
			//bmp.write( (const char *)p_bmpData + y * p_width * 3,
			//	p_width * 3);
			if( paddingLength > 0) {
				bmp.write( (char *)padding, paddingLength);
			}
		}

		/* Determine the size of the file after the image data has
		 * been written. Also, any padding bytes applied have to
		 * be considered as well.
		 */
		fileSize += (p_width * 3) * p_height;
		fileSize += paddingLength * p_height;
		break;

	}

	/*******************************************************************
	 * Write out the bitmap file header.
	 *******************************************************************/
	bmp.seekp( 0, std::ios::beg);
	bmp.write( "BM", 2);
	SET_AND_WRITE_DATA( dwordData, fileSize);
	SET_AND_WRITE_DATA( wordData, 0);
	SET_AND_WRITE_DATA( wordData, 0);
	SET_AND_WRITE_DATA( dwordData, imageDataOffset);

	return RGERR_SUCCESS;
}

int rgBMP::generate_bmp( const int w, const int h, const int p_depth)
{
	clear();
	if( (w <= 0) || (h <= 0)) return -1;

	
	switch( p_depth) {
	// XXX: Thanks to "ref_0" for the info on why typedefs should be used.
	case 8: imageData = new rgBYTE[ w * h]; break;
	case 16: imageData = new rgBYTE[ w * h * 2]; break;
	case 24: imageData = new rgBYTE[ w * h * 3]; break;
	default: return -2;
	}
	if( !imageData) return -3;
	width = w;
	height = h;
	colorDepth = p_depth;
	return 0;
}

int rgBMP::copy_sub_image( rgBMP &src, int srcX, int srcY, int desX, int desY,
	int p_width, int p_height)
{
	if( &src == this) return -1;
	if( (srcX < 0) || (srcX >= src.getWidth())) return -2;
	if( (srcY < 0) || (srcY >= src.getHeight())) return -3;
	if( (srcX + p_width) > src.getWidth()) return -4;
	if( (srcY + p_height) > src.getHeight()) return -5;

	if( (desX < 0) || (desX >= getWidth())) return -6;
	if( (desY < 0) || (desY >= getHeight())) return -7;
	if( (desX + p_width) > getWidth()) return -8;
	if( (desY + p_height) > getHeight()) return -9;

	if( src.getColorDepth() != getColorDepth()) return -10;

	int span, src_span, des_span;
	int srcXOffset;
	int desXOffset;
	switch( getColorDepth()) {
	case 8:
		span = p_width;
		src_span = src.getWidth();
		des_span = getWidth();
		srcXOffset = srcX;
		desXOffset = desX;
		break;

	case 16:
		span = p_width * 2;
		src_span = src.getWidth() * 2;
		des_span = getWidth() * 2;
		srcXOffset = srcX * 2;
		desXOffset = desX * 2;
		break;

	case 24:
		span = p_width * 3;
		src_span = src.getWidth() * 3;
		des_span = getWidth() * 3;
		srcXOffset = srcX * 3;
		desXOffset = desX * 3;
		break;

	default: return -11;
	}

	const rgBYTE *src_img = src.getImageData() + (srcY * src_span)
		+ srcXOffset;
	rgBYTE *des_img = getImageData() + (desY * des_span)
		+ desXOffset;
	for( int y = 0; y < p_height; y++) {
		memcpy( des_img, src_img, span);
		src_img += src_span;
		des_img += des_span;
	}
	return 0;
}










