/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * rgSDK/rgShadow/rgShadow.h
 *
 * Copyright (C) 2004 Amour Hassan
 *
 * XXX: Some stencil shadowing information taken from below address:
 * http://www.opengl.org/resources/tutorials/advanced/advanced96/node48.html
 *
 *
 * TODO: 1. Implement Z-Fail stencil shadowing.
 *	 2. Implement ability to render in and outside of shadow volumes.
 *	 3. Determine effect of rendering within multiple shadow volumes.
 *
 *
 * XXX: Z-Pass Stencil Shadowing
 *-----------------------------------------------------------------------------
 *
 *	1. Enable the color buffer for writing.
 *		example: glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
 *
 *	2. Enable the depth buffer for writing.
 *		example: glEnable( GL_DEPTH_TEST);
 *			 glDepthMask( GL_TRUE);
 *
 *	3. Prepare scene for rendering in shadow (i.e. Turn off light source).
 *		example: glDisable( GL_LIGHTING);
 *
 *			OR
 *
 *		example: glColor3f( 0.2, 0.2, 0.2);	// Darken scene
 *
 *	5. Render the entire scene.
 *
 *	6. Disable writing to the colour and depth buffers.
 *		example: glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
 *			 glDepthMask( GL_FALSE);
 *
 *	7. Clear the stencil buffer to 0 if the eye is outside the shadow
 *	   volume, or 1 if inside.
 *		example: glClearStencil( 0);	// If outside
 *
 *			OR
 *
 *		example: glClearStencil( 1);	// If inside
 *
 *	8. Set the stencil function to always pass.
 *		example: glStencilFunc( GL_ALWAYS, 0, 0xFFFFFFFF);
 *
 *	9. Set the stencil operations to increment if the depth test passes.
 *		example: glStencilOp( GL_KEEP, GL_KEEP, GL_INCR);
 *
 *	10. Turn on back face culling (by default, GL_CCW (counter clockwise) is
 *	   the front face. Use "glFrontFace()" to change to GL_CW (clockwise
 *	   if neccasary).
 *		example: glEnable( GL_CULL_FACE);
 *			 glCullFace( GL_BACK);
 *
 *	11. Render the shadow volume polygons.
 *
 *	12. Set the stencil operations to decrement if the depth test passes.
 *		example: glStencilOp( GL_KEEP, GL_KEEP, GL_DECR);
 *
 *	13. Turn on front face culling.
 *		example: glCullFace( GL_FRONT);
 *
 *	14. Render the shadow volume polygons.
 *
 *	15. Set the stencil function to test for equality to 0.
 *		example: glStencilFunc( GL_EQUAL, 0, 0xFFFFFFFF);
 *
 *	16. Set the stencil operations to do nothing.
 *		example: glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP);
 *
 *	17. Turn on the light source.
 *		example: glEnable( GL_LIGHTING);
 *
 *			OR
 *
 *		example: glColor3f( 1.0, 1.0, 1.0);	// Lighten scene
 *
 *	18. Render the entire scene. 
 *
 */

#ifndef RGSHADOW_H
#define RGSHADOW_H

#include <rgStandard.h>

class sk3DRomModel;

struct rgShadowQuad
{
	rgVector v[4];
};

struct rgShadowTri
{
	int casts_shadow;
	int vid[3];
	rgVector normal;
};

struct rgShadowTriCap
{
	rgVector v[3];
};

#define RGSHADOW_VOLUME_V0V1_ORDER	0
#define RGSHADOW_VOLUME_V1V2_ORDER	1
#define RGSHADOW_VOLUME_V2V0_ORDER	2
struct rgShadowEdgeEntry
{
	char used;	// If the edge has been used.
	char order;
	// order is one of:
	//	RGSHADOW_VOLUME_V0V1_ORDER or
	//	RGSHADOW_VOLUME_V1V2_ORDER or
	//	RGSHADOW_VOLUME_V2V0_ORDER
	short tid;	// Which triangle owns the edge
};

class rgShadowVolume
{
public:

	rgShadowVolume();
	~rgShadowVolume();
	void clear();

	// Description: Creates a shadow from the given geometry. The shadow
	//	created is not renderable until it is "cast" using one of the
	//	casting functions. Once the shadow has been cast, it can be
	//	rendered.
	void create( const sk3DRomModel *mdl);

	void cast_ortho( const rgVector w) {
		if( !ready_for_use()) return;
		build_ortho_silhouette( w);
	}

	void cast_projected( const rgVector p) {
		if( !ready_for_use()) return;
		build_projected_silhouette( p);
	}

	bool ready_for_use() const {
		if( !m_caster) return false;
		if( !m_vert) return false;
		if( !m_tri) return false;
		if( !m_edge_table) return false;
		if( !m_quad) return false;
		return true;
	}

	void set_shadow_cast_length( const rgFloat len) {
		m_shadow_cast_length = fabs( len);}

	void render() const;

	int tri_count() const { return m_tcount;}
	int quad_count() const { return m_qcount;}

	// Description: Call this function when the model has changed (such
	//	as being transformed) so that that its shadow volume will be
	//	updated as well. This function is automatically called when
	//	the "create" function is called.
	void update_vertices();
	void update_triangle_normals();

private:

	const sk3DRomModel *m_caster;

	rgFloat m_shadow_cast_length;

	rgVector *m_vert;
	int m_vcount;

	// Edge table is used for determining how (and what) quads are
	// generated to produce the shadow volume.
	rgShadowEdgeEntry *m_edge_table;
	int m_esize;

	rgShadowTri *m_tri;
	int m_tcount;

	// The quads define the shadow volume. The size of the "m_quad" array
	// is exactly (m_tcount * 3). This is because "m_tcount * 3" is the
	// maximum possible number of shadow quads that can be produced.
	rgShadowQuad *m_quad;

	rgShadowTriCap *m_tri_cap;
	int m_tri_cap_count;

	// Total number of renderable shadow quads.
	int m_qcount;

	int create_vertex_list();
	int create_tri_list();
	int create_quad_list();
	int create_edge_table();
	int create_edge_list();
	int create_tri_cap_list();

	void clear_vertex_list() {
		if( m_vert) { delete [] m_vert; m_vert = 0;}
		m_vcount = 0;
	}

	void clear_tri_list() {
		if( m_tri) { delete [] m_tri; m_tri = 0;}
		m_tcount = 0;
	}

	void clear_edge_table() {
		if( m_edge_table) { delete [] m_edge_table; m_edge_table = 0;}
		m_esize = 0;
	}

	void clear_quad_list() {
		if( m_quad) { delete [] m_quad; m_quad = 0;}
		m_qcount = 0;
	}

	void clear_tri_cap_list() {
		if( m_tri_cap) { delete [] m_tri_cap; m_tri_cap = 0;}
		m_tri_cap_count = 0;
	}

	void build_ortho_silhouette( const rgVector w);
	void build_ortho_quad_list( const rgVector w);
	void build_ortho_capping_triangles( const rgVector w);
	void build_ortho_quad( const rgShadowEdgeEntry *e, rgShadowQuad *q,
		const rgVector w);

	void build_projected_silhouette( const rgVector p);
	void build_projected_quad_list( const rgVector p);
	void build_projected_capping_triangles( const rgVector p);
	void build_projected_quad( const rgShadowEdgeEntry *e, rgShadowQuad *q,
		const rgVector w);

	void mark_triangle_edges_as_used( rgShadowTri *t, const int tid);	
	void build_quad( const rgShadowEdgeEntry *e, rgShadowQuad *q)
		const;

	void mark_edge_as_used( int a, int b, const int tid, const int order);

	void render_quads_using_triangles() const;
	void render_capping_triangles() const;
};

#endif
