/*
Copyright (C) 1999-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RG_STANDARD_H
#define RG_STANDARD_H

#include <iostream>
#include <rgPort.h>
#include <math.h>

#define RG_NULL 0

/* This is the minmum value that complex floating point matrix operations will
 * tolerate. This is especially true for matrix inversion. If any element in
 * the matrix is less then RG_MAT_MIN_FLOAT_VALUE, then it will be treated as
 * a zero. This is not an ideal solution but it "should" work for "reasonable"
 * matrix operations. Any code that uses this minimum value is most likely
 * unrealiable.
 *
 * IMPORTANT NOTE: The value of RG_MAT_MIN_FLOAT_VALUE was "carefully"
 *	selected and cannot be truely arbitrary.
 */
#define RG_MAT_MIN_FLOAT_VALUE 0.000005f

/* Common radian angles */
#define RG_45_DEG_IN_RAD 0.785398163f
#define RG_90_DEG_IN_RAD 1.570796327f
#define RG_180_DEG_IN_RAD 3.141592654f
#define RG_270_DEG_IN_RAD 4.71238898f
#define RG_360_DEG_IN_RAD 6.283185308f

#define RG_PI RG_180_DEG_IN_RAD
#define RG_2PI RG_360_DEG_IN_RAD

#define RG_INV_45_DEG_IN_RAD	1.273239545f

/* Conversion from degrees to radians and vice versa */
#define RG_PI_OVER_180 0.017453293f
#define RG_180_OVER_PI 57.29577951f
#define RG_DEGTORAD(d) ((d)*RG_PI_OVER_180)
#define RG_RADTODEG(r) ((r)*RG_180_OVER_PI)

/* Determing minimum and maximum values */
#define RG_MIN( a, b) (((a) <= (b)) ? (a) : (b))
#define RG_MAX( a, b) (((a) >= (b)) ? (a) : (b))

/*
 * Description: Performs a swap of the parameters "a" and "b".
 */
template <class T>
void rgSwap( T &a, T &b) { T c = a; a = b; b = c;}


template <class T>
class rgTempArray
{
public:
	T *data;

	rgTempArray() { init();}
	rgTempArray( const int count) {
		init();
		create( count);
	}
	~rgTempArray() { free();}

	bool create( const int count) {
		rgTempArray::free();
		if( count <= 0) return false;
		if( !(data = new T[ count])) return false;
		m_size = count;
		return true;
	}
	void free() {
		if( data) { delete [] data; data = 0;}
		m_size = 0;
	}
	int size() const { return m_size;}
	bool empty() const {
		if( m_size > 0) return false;
		else return true;
	}

	T *detach() {
		T *tmp = data;
		data = 0;
		m_size = 0;
		return tmp;
	}

protected:
	int m_size;
	void init() {
		data = 0;
		m_size = 0;
	}
};

/*! \brief A 3D vector class providing basic vector operations.
  *
  * A 3D vector class providing basic vector operations.
  */
class rgVector
{
public:

	/* NOTE: THE SIZE OF THIS CLASS MUST BE 3*sizeof(rgFLoat) */
	rgFloat x;
	rgFloat y;
	rgFloat z;
	/* NOTE: NO MORE VARIABLE MEMBERS ALLOWED */

	rgVector();
	rgVector( rgFloat p_x, rgFloat p_y, rgFloat p_z);
    
	friend std::ostream &operator <<( std::ostream &o, rgVector v);

	bool operator ==( rgVector v) const {
		if( (x == v.x) && (y == v.y) && (z == v.z)) return true;
		return false;
	}

	rgVector &operator +=( const rgVector v) {
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

	rgVector &operator -=( const rgVector v) {
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}

	friend rgVector operator -( rgVector v) {
		v.x = -v.x; v.y = -v.y; v.z = -v.z; return v;}
        
	rgVector operator +( rgVector v) const {
		v.x += x; v.y += y; v.z += z; return v;}

	rgVector operator -( rgVector v) const {
		v.x = x - v.x; v.y = y - v.y; v.z = z - v.z; return v;}

	rgVector operator *( rgFloat k) const {
		rgVector u;
		u.x = x * k; u.y = y * k; u.z = z * k; return u;}

	friend rgVector operator *( const rgFloat k, rgVector v) {
		v.x *= k; v.y *= k; v.z *= k; return v;}

	rgFloat length_squared() const { return (x*x + y*y + z*z);}
	rgFloat length() const { return (rgFloat)sqrt( length_squared());}
    
	rgVector &normalize();
   
	rgVector cross( const rgVector v) const {
		return rgVector( y * v.z - z * v.y, z * v.x - x * v.z,
			x * v.y - y * v.x);
	}
	rgVector cross( const rgFloat p_x, const rgFloat p_y,
		const rgFloat p_z) const {
		return cross( rgVector( p_x, p_y, p_z));
	}
		
	rgFloat dot( const rgVector v) const {
		return (x * v.x) + (y * v.y) + (z * v.z);}

	bool null() const {
		if( (0.0f == x) && (0.0f == y) && (0.0f == z)) return true;
		return false;
	}

	// XXX: Reads in little endian vector.
	rgVector &read_binary( std::istream &src) {
		x = read_float( src);
		y = read_float( src);
		z = read_float( src);
		return *this;
	}

	// XXX: Vector is written out in little endian format.
	void write_binary( std::ostream &des) const {
		write_float( des, x);
		write_float( des, y);
		write_float( des, z);
	}

};

typedef rgVector rgVector3D;

/*
class rgHVector : public rgVector
{
public:
	float w;

	rgHVector();
	rgHVector( float p_x, float p_y, float p_z);
	rgHVector( float p_x, float p_y, float p_z, float p_w);
	rgHVector( rgVector3D v);

	rgVector4D operator +( rgVector4D v);
	rgVector4D operator -( rgVector4D v);
	rgVector4D operator *( const rgFloat k);
	friend rgVector operator *( const rgFloat k, rgVector4D v);
	rgFloat length();
	rgFloat normalize();
	rgFloat dot( rgVector4D);
}

typedef rgHVector rgVector4D;
*/

class rgCoordSystem;

/*! \brief A matrix class providing basic matrix operations.
 */
class rgMatrix
{
public:

	/*! \brief The elements of the matrix.
	 */
	rgFloat d[16];		/* 4 x 4 matrix */
    
	/*! \brief Constructs a new identity matrix.
	 */
	rgMatrix();

	rgMatrix( const rgCoordSystem &cs);
	rgMatrix &operator =( const rgCoordSystem &cs) {
		convert_csys_to_matrix( cs);
		return *this;
	}

	rgMatrix &apply_positioning_csys( const rgCoordSystem &cs);

	/*! \brief Matrix/Vector multiplication.
	 */
	rgVector operator *( const rgVector v) const {
		return rgVector(
			d[0] * v.x + d[1] * v.y + d[2] * v.z,
			d[4] * v.x + d[5] * v.y + d[6] * v.z,
			d[8] * v.x + d[9] * v.y + d[10] * v.z);
	}
    
	/*! \brief Matrix addition.
	 */
	rgMatrix operator +( const rgMatrix &m) const;
    
	/*! \brief Matrix subtraction.
	 */
	rgMatrix operator -( const rgMatrix &m) const;

	/*! \brief Matrix multiplication ("this" * "m").
	 */
	rgMatrix operator *( const rgMatrix &m) const {
		rgMatrix n( *this);
		return (n *= m);	// XXX: MUST use "*=" not "*"
	}

	rgMatrix &operator +=( const rgMatrix &m);
	rgMatrix &operator -=( const rgMatrix &m);
	rgMatrix &operator *=( const rgFloat);
	rgMatrix &operator *=( const rgMatrix &m);

	friend std::ostream &operator <<( std::ostream &o, const rgMatrix &m);

	/*! \brief Loads a 4 x 4 identity matrix.
	 */
	void loadIdentity();

	rgVector pointMul( const rgVector v) const {
		return pointMul( v.x, v.y, v.z);}

	rgVector pointMul( const rgFloat x, const rgFloat y, const rgFloat z)
		const
	{
		return rgVector(
			d[0] * x + d[1] * y + d[2] * z + d[3],
			d[4] * x + d[5] * y + d[6] * z + d[7],
			d[8] * x + d[9] * y + d[10] * z + d[11]);
	}

	void rotate( const rgFloat ang, const rgVector v) {
		rotate( ang, v.x, v.y, v.z);}
	void rotate( const rgFloat ang, const rgFloat x, const rgFloat y,
		const rgFloat z);
	
	void scale( rgVector v);
	void scale( const rgFloat x, const rgFloat y, const rgFloat z);
	
	void translate( rgVector v);
	void translate( const rgFloat x, const rgFloat y, const rgFloat z);

	void transpose() {
		rgSwap( d[1], d[4]);
		rgSwap( d[2], d[8]);
		rgSwap( d[3], d[12]);
		rgSwap( d[6], d[9]);
		rgSwap( d[7], d[13]);
		rgSwap( d[11], d[14]);
	}
	
	bool frustum( rgFloat left, rgFloat right, rgFloat bottom, rgFloat top,
		rgFloat zNear, rgFloat zFar);
	bool ortho( rgFloat left, rgFloat right, rgFloat bottom, rgFloat top,
		rgFloat zNear, rgFloat zFar);
	/*
	 * Parameters:
	 * 	fovy - Field of view in degrees. Range is 1 to 180 degrees.
	 */
	bool perspective( rgFloat fovy, rgFloat aspect, rgFloat zNear,
		rgFloat zFar);
	bool perspective( rgFloat fovy, rgFloat aspect, rgFloat zNear,
		rgFloat zFar, rgFloat *p_width, rgFloat *p_height);

	/* IMPORTANT NOTE: invert() is an unreliable function because it makes
	 *	use of (or is dependent on functions which make use of)
	 *	RG_MAT_MIN_FLOAT_VALUE.
	 */
	bool invert();

	/* IMPORTANT NOTE: Any code which has to use this function is most
	 *	likely unreliable. This function forces any matrix elements
	 *	that are less then RG_MAT_MIN_FLOAT_VALUE to be 0.0f.
	 *
	 *	It is recommended that the user of the rgMatrix class not
	 *	use this function directly. That is, it should only be used
	 *	internally.  However it is made public so that it can be used
	 *	to make debugging outputs more readable by making values
	 *	which are virtualy 0.0f, equal 0.0f. Once again, this function
	 *	should only be called directly for rgMatrix class debugging
	 *	only and is not for normal use.
	 *
	 * TODO: precisionGaurd should be removed as soon as a suitable
	 *	high precision calculation method is available.
	 */
	void precisionGaurd();

private:

	bool rowElementLeads( rgFloat *d, int col);
	bool makeLeadingOne( const int tarRow, const int tarCol,
		rgMatrix &inv);
	void convert_csys_to_matrix( const rgCoordSystem &cs);
};

class rgBasis
{
public:
	rgVector up;
	rgVector side;
	rgVector eye;

	rgBasis();
	friend std::ostream &operator <<( std::ostream &o, rgBasis &b);
};

class rgCoordSystem : public rgBasis
{
public:
	rgVector p;	// Origin position

	friend std::ostream &operator <<( std::ostream &o,
		const rgCoordSystem &b);

	bool operator ==( const rgCoordSystem &cs) const {
		if(	(p == cs.p) && (side == cs.side) &&
			(up == cs.up) && (eye == cs.eye)) return true;
		return false;
	}

	rgMatrix get_mapping_matrix() const {
		rgMatrix m( *this);
		rgSwap( m.d[1], m.d[4]);
		rgSwap( m.d[2], m.d[8]);
		rgSwap( m.d[6], m.d[9]);
		m.d[3] = 0;
		m.d[7] = 0;
		m.d[11] = 0;
		rgMatrix t;
		t.translate( p);
		return t * m;
	}

	// Description: Maps a given point such that it is relative to "this"
	//	coordinate system.
	// NOTE: Points are affected by position.
	rgVector map_point( const rgVector q) const {
		return map_vector( q) + p;}

	// Descriotion: Maps a given vector such that it is relative to "this"
	//	coordinate system.
	// NOTE: Vectors are not affected by position.
	rgVector map_vector( const rgVector v) const {
		return (v.x * side) + (v.y * up) + (v.z * eye);}

	// Description: Maps a given coordinate system such that it is relative
	//	to "this" coordinate system.
	void map_coord_system( rgCoordSystem &dcs, rgCoordSystem &scs);
	rgCoordSystem map_csys( const rgCoordSystem p_cs) const {
		rgCoordSystem cs;
	        cs.p = map_point( p_cs.p);
	        cs.side = map_vector( p_cs.side);
	        cs.up = map_vector( p_cs.up);
	        cs.eye = map_vector( p_cs.eye);
	        return cs;}

	//rgVector change_point_csys( rgVector point, rgCoordSystem &old);
	//rgVector change_vector_csys( rgVector v, rgCoordSystem &old);
	rgVector change_point_csys( const rgVector q) const {
		return change_vector_csys( q - p);}
	rgVector change_vector_csys( const rgVector v) const {
		return  rgVector( side.dot( v), up.dot( v), eye.dot( v));}
	rgCoordSystem change_csys_csys( const rgCoordSystem cs) const;

	/*
	rgCoordSystem &transform( const rgMatrix &m) {
		p = m.pointMul( p);
		side = m * side;
		up = m * up;
		eye = m * eye;
		return *this;
	}
	*/

	rgCoordSystem &translate( const rgVector v) {
		rgMatrix t;
		t.translate( v);
		return translate( t);
	}
	rgCoordSystem &translate( const rgMatrix &t) {
		p = t.pointMul( p);
		return *this;
	}

	rgCoordSystem &rotate( const float ang, float x, float y, float z) {
		return rotate( ang, rgVector( x, y, z));
	}
	rgCoordSystem &rotate( const float ang, const rgVector axis) {
		rgMatrix m;
		m.rotate( ang, axis);
		return rotate( m);
	}
	rgCoordSystem &rotate( const rgMatrix &m) {
		side = m * side;
		up = m * up;
		eye = m * eye;
		return *this;
	}

	rgCoordSystem &scale( float x, float y, float z) {
		return scale( rgVector( x, y, z));}
	rgCoordSystem &scale( const rgVector v) {
		side = side * v.x;
		up = up * v.y;
		eye = eye * v.z;
		return *this;
	}

	void write_binary( std::ostream &des) const {
		p.write_binary( des);
		side.write_binary( des);
		up.write_binary( des);
		eye.write_binary( des);
	}

	rgCoordSystem &read_binary( std::istream &src) {
		p.read_binary( src);
		side.read_binary( src);
		up.read_binary( src);
		eye.read_binary( src);
		return *this;
	}
};

/*
 * Description: Returns the angle between vector "a" and vector "b". The angle
 *	returned is in radians and will range from 0 to 0.5 * PI.
 */
rgFloat rgDotAngle( const rgVector a, const rgVector b);

/*
 * Description: Returns the angle that the vector (p - origin) makes with
 * 	positive X axis (X axis is positive going right). The angle returned
 *	can range from 0 to 2 * PI. Angles returned are in radians.
 */
rgFloat rgAngle360( const rgVector origin, rgVector p);

int sphereRayIntersection( rgVector sCenter, rgFloat sRadius, rgVector rp,
        rgVector rd, rgVector &cp1, rgVector &cp2);

inline rgVector rgProjectAlong( const rgVector v, const rgVector u) {
	const rgFloat lengthSquared = u.length_squared();
	if( 0.0f == lengthSquared) return rgVector();
	return (u.dot( v) / lengthSquared) * u;
}

inline rgVector rgReflectAbout( const rgVector v, const rgVector n) {
	return v + (2.0f * (rgProjectAlong( v, n) - v));
}

rgFloat rgPointPlaneDis( rgVector &pp, rgVector &pn, rgVector &p);

bool rgRayPlaneIntersect( rgVector pp, rgVector pn, rgVector rp, rgVector rd,
	rgVector &ip);

bool rgLineSegPlaneIntersect( rgVector pp, rgVector pn, rgVector a,
	rgVector b, rgVector &ip);

bool rgLineSegPlaneIntersect( rgVector pp, rgVector pn, rgVector a,
	rgVector b, rgVector &ip, rgFloat &t);

bool rgLinePlaneIntersect( rgVector pp, rgVector pn, rgVector lp, rgVector ld,
        rgVector &ip);

inline bool rgLinePlaneIntersectScalar( rgVector &pp, rgVector &pn,
	rgVector &lp, rgVector &ld, rgFloat &t);

inline bool rgPlanarPointInTriangle( const rgVector a, const rgVector b,
	const rgVector c, const rgVector p) {
	const rgVector n( (b - a).cross( p - a));
	if( n.dot( (c - b).cross( p - b)) <= 0.0f) return false;
	if( n.dot( (a - c).cross( p - c)) <= 0.0f) return false;
	return true;
}

inline bool rgPlanarPointInQuad( const rgVector a, const rgVector b,
	const rgVector c, const rgVector d, const rgVector p) {
	const rgVector n( (b - a).cross( p - a));
	if( n.dot( (c - b).cross( p - b)) <= 0.0f) return false;
	if( n.dot( (d - c).cross( p - c)) <= 0.0f) return false;
	if( n.dot( (a - d).cross( p - d)) <= 0.0f) return false;
	return true;
}

inline bool rgPlanarPointInHexagon( const rgVector *v, const rgVector p) {
	const rgVector n( (v[1] - v[0]).cross( p - v[0]));
	if( n.dot( (v[2] - v[1]).cross( p - v[1])) <= 0.0f) return false;
	if( n.dot( (v[3] - v[2]).cross( p - v[2])) <= 0.0f) return false;
	if( n.dot( (v[4] - v[3]).cross( p - v[3])) <= 0.0f) return false;
	if( n.dot( (v[5] - v[4]).cross( p - v[4])) <= 0.0f) return false;
	if( n.dot( (v[0] - v[5]).cross( p - v[5])) <= 0.0f) return false;
	return true;
}


inline bool rgPlanarPointInQuad( const rgVector *v, const rgVector p) {
	return rgPlanarPointInQuad( v[0], v[1], v[2], v[3], p);}

//bool rgPlanarPointInConvexFace( const rgVector *f, const int count,
//	const rgVector p);
inline bool rgPlanarPointInConvexFace( const rgVector *f, const int count,
        const rgVector p)
{
	if( count < 3) return false;
	const rgVector bn( (f[1] - f[0]).cross( p - f[0]));
	for( int i = 1; i < (count - 1); i++) {
		const rgVector n( ( f[i+1] - f[i]).cross( p - f[i]));
		if( bn.dot( n) <= 0.0f) return false;
	}

	const rgVector n( ( f[0] - f[count - 1]).cross( p - f[count - 1]));
	if( bn.dot( n) <= 0.0f) return false;
	return true;
}


bool rgRayTriangleIntersect( rgVector a, rgVector b, rgVector c, rgVector n,
        rgVector rp, rgVector rd, rgVector &ip);

bool rgLineLineIntersect( rgVector a, rgVector b, rgVector c, rgVector d,
	rgVector &p);

bool rgRayLineSegIntersect( rgVector rp, rgVector rd, rgVector c, rgVector d,
	rgVector &p);

bool rgRayLineSegIntersect( rgVector rp, rgVector rd, rgVector c, rgVector d,
	rgVector &p, rgFloat &t, rgFloat &k);


inline bool rgLineSegLineSegIntersect( const rgVector a, const rgVector b,
	const rgVector c, const rgVector d, rgVector &p, rgFloat &t, rgFloat &k)
{
	const rgVector ba( b - a);
	const rgVector ca( c - a);

	// Calculate a "normal-like" vector.
	const rgVector n( ca - rgProjectAlong( ca, ba));

	if( 0.0f == ca.dot( n)) {
		t = 0.0f;
		p = c;
	
	} else {
		const rgFloat dc_dot_n = (d - c).dot( n);
		if( 0.0f == dc_dot_n) return false;

		t = (a.dot( n) - c.dot( n)) / dc_dot_n;
		if( (t < 0.0f) || (t > 1.0f)) return false;

		p = c + ((d - c) * t);
	}

	const rgFloat ba_dot_ba = ba.dot( ba);
	if( 0.0f == ba_dot_ba) {
		if( p == a) {
			k = 0.0f;
			return true;
		}
		return false;
	}

	k = (p - a).dot( ba) / ba_dot_ba;
	if( (k < 0.0f) || (k > 1.0f)) return false;
	return true;
}

// XXX: "a" must NOT == "b". If "a" == "b", then it is not a line segment,
//	it is a point (which is not allowed).
inline bool rgLineSegLineSegIntersect( const rgVector a, const rgVector b,
	const rgVector c, const rgVector d)
{
	const rgVector ba( b - a);
	const rgVector ca( c - a);

	// Calculate a "normal-like" vector.
	const rgVector n( ca - rgProjectAlong( ca, ba));

	float k;
	if( ca.dot( n) != 0.0f) {
		const rgFloat dc_dot_n = (d - c).dot( n);
		if( 0.0f == dc_dot_n) return false;
		const float t = (a.dot( n) - c.dot( n)) / dc_dot_n;

		if( (t < 0.0f) || (t > 1.0f)) return false;
		const rgVector p( c + ((d - c) * t));
		k = (p - a).dot( ba) / ba.dot( ba);
	
	} else k = ca.dot( ba) / ba.dot( ba);

	if( (k < 0.0f) || (k > 1.0f)) return false;
	return true;
}

inline bool rgLineSegLineSegIntersect( rgVector a, rgVector b, rgVector c,
        rgVector d, rgVector &p)
{
	rgFloat t, k;
	return rgLineSegLineSegIntersect( a, b, c, d, p, t, k);
}

bool rgLineSegLineSegIntersect( rgVector a, rgVector b, rgVector c, rgVector d,
	rgVector &p, rgFloat &t, rgFloat &k);

int rgCircleRayIntersect( rgVector cp, const rgFloat r, rgVector rp,
	rgVector rd, rgVector &ip1, rgVector &ip2);

inline rgVector rgMakePlanar( const rgVector v, const rgVector normal) {
	return (v - (v.dot( normal) * normal));};

inline rgVector rgCalcTriangleNormal( const rgVector a, const rgVector b,
	const rgVector c) { return ((b - a).cross( a - c));}

rgVector rgMapToBasis( rgVector p, rgBasis &ob, rgBasis &nb);

inline int rg_summation( const int n) {
	if( n <= 0) return 0;
	return ((1 + n) * n) >> 1;
}

inline int rg_lower_triangle_array_access( const int row, const int col) {
	return rg_summation( row - 1) + col;
}

inline int rg_upper_triangle_array_access( const int row, const int col,
	const int row_width) {
	return ((row * row_width) - rg_summation( row - 1)) + col;
}

/**************************************************************************
 * Inline functions begin
 **************************************************************************/
inline void rgVector_cross( rgVector &n, rgVector &u, rgVector &v);
inline rgFloat rgVector_dot( rgVector &u, rgVector &v);
inline void rgVector_scalar_mul( rgVector &u, const rgFloat k, rgVector &v);
/**************************************************************************
 * Inline functions end
 **************************************************************************/

#endif
