/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "engine_game.h"
#include "engine_physics.h"
#include "engine_graphics.h"
#include <rgOpenGL.h>
#include <SDL/SDL.h>
#include <GL/glu.h>
#include <rg3DRom.h>
#include <rgError.h>
#include "things/things.h"

void gmCamera::set_owner( ThingControlBlock *p)
{
	if( m_owner && (!(m_owner->is_free())))
		m_owner->m_thing->uncontrol_camera();
	m_owner = p;
}

GraphicsEngine::GraphicsEngine( GameEngine *ge)
{
	game_engine = ge;

	m_enabled = true;

	m_draw_sector_level = 32768;
	m_sector_line_width = 1;
	m_draw_sectors = false;

	m_model_man = new ModelManager();
	root_renderer = 0;
}

GraphicsEngine::~GraphicsEngine()
{
	clear();
	const char tmp_emsg_head[] = "~GraphicsEngine(): Error ";
	while( !(m_camera.empty())) {
		std::map <const std::string, gmCamera *> :: iterator iter =
			m_camera.begin();
		const std::string key( iter->first);
		const gmCamera *cam = iter->second;
		const int err = remove_camera( iter);
		if( err) {
			std::cerr << (tmp_emsg_head + gm_to_string( err) +
				" while removing camera \"" + key +
				"\" ( " + gm_to_string( cam) + ")\n"
				"Canceling camera delete ( " +
				gm_to_string( int( m_camera.size())) +
				" camera(s) remain.)\n");
			break;
		}
	}

	while( !(m_texture.empty())) {
		std::map <const std::string, TextureHandle *> :: iterator iter
			= m_texture.begin();
		const std::string key( iter->first);
		const TextureHandle *th = iter->second;
		const int err = delete_texture( key);
		if( err) {
			std::cerr << (tmp_emsg_head + gm_to_string( err) +
				" while deleting texture \"" + key + "\" (" +
				gm_to_string( th) + ")\n"
				"Canceling texture delete ( " +
				gm_to_string( int( m_texture.size())) +
				" texture(s) remain.)\n");
			break;
		}
	}

	if( m_model_man) delete m_model_man;
}

void GraphicsEngine::startup()
{
	init_gl_state();
	set_projection();
}

void GraphicsEngine::clear()
{
	reset_all_cameras();

	init_gl_state();
	set_projection();
}

void GraphicsEngine::reset_all_cameras()
{
	std::map <const std::string, gmCamera  *> :: const_iterator iter =
		m_camera.begin();
	for( ; iter != m_camera.end(); iter++) {
		(iter->second)->cs.p = rgVector( 0, 100, 0);
		(iter->second)->look( rgVector( 0, 0, 1),
			rgVector( 0, 1, 0));
	}
}

bool GraphicsEngine::set_current_camera( const std::string &key)
{
	if( get_camera( key)) m_current_camera = key;
	else return false;
	return true;
}

int GraphicsEngine::add_camera( const std::string &key)
{
	return add_camera( key, rgVector(), rgVector( 0, 0, 1),
		rgVector( 0, 1, 0));
}

int GraphicsEngine::add_camera( const std::string &key,
	rgVector pos, rgVector p_look_at)
{
	rgGLCamera tmp;
	tmp.cs.p = pos;
	tmp.look_at( p_look_at);
	return add_camera( key, pos, p_look_at, tmp.cs.up);
}

int GraphicsEngine::add_camera( const std::string &key,
	rgVector pos, rgVector p_look_at, rgVector up)
{
	gmCamera *p = new gmCamera( pos, p_look_at, up);
	if( !p) return -1;
	typedef std::pair <const std::string, gmCamera *> SCPair;
	if( false == ((m_camera.insert( SCPair( key, p))).second)) {
		delete p;
		return -2;
	}
	return 0;
}

int GraphicsEngine::remove_camera( const std::string &key)
{
	std::map <const std::string, gmCamera *> :: iterator iter =
		m_camera.find( key);
	return remove_camera( iter);
}

int GraphicsEngine::remove_camera( std::map <const std::string, gmCamera *> ::
	iterator &iter)
{
	if( iter == m_camera.end()) return -1;
	if( (iter->second)->owner()) return -2;
	if( iter->second) delete (iter->second);
	m_camera.erase( iter);
	return 0;
}

std::string hlp_list_camera(
	std::map <const std::string, gmCamera *> :: const_iterator &iter)
{
	std::string tmp( gm_to_string( iter->second) + ",\"" + (iter->first));
	if( (iter->second)->owner()) {
		tmp += "\" owner = \"" +
			((iter->second)->owner())->m_key + "\"\n";
	} else tmp += "\" owner = none\n";
	return tmp;
}
std::string GraphicsEngine::list_cameras() const
{
	std::string tmp( "Available Cameras (pointer, key, owner):\n");
	std::map <const std::string, gmCamera  *> :: const_iterator iter =
		m_camera.begin();
	for( ; iter != m_camera.end(); iter++)
		tmp += "\t" + hlp_list_camera( iter);
	tmp += "Total Cameras: " + gm_to_string( int( m_camera.size())) + "\n";
	return tmp;
}

std::string GraphicsEngine::dump_camera( const std::string &key) const
{
	std::map <const std::string, gmCamera *> :: const_iterator iter =
		m_camera.find( key);
	if( iter == m_camera.end())
		return std::string( "no such camera \"" + key + "\"\n");
	return dump_camera( iter);
}

std::string GraphicsEngine::dump_camera(
	std::map <const std::string, gmCamera *> :: const_iterator &iter)
	const
{
	std::string tmp( "Camera \"" + (iter->first) + "\", " +
		gm_to_string( iter->second) + "\n");
	tmp += "\towner = ";
	if( (iter->second)->owner()) {
		tmp += "\"" + ((iter->second)->owner())->m_key +
			"\"\n";
	} else tmp += "none\n";
	
	gmCamera *c = (iter->second);
	tmp +=	"\tposition = " + gm_to_string( c->cs.p) + "\n" +
		"\tside = " + gm_to_string( c->cs.side) + "\n" +
		"\tup = " + gm_to_string( c->cs.up) + "\n" +
		"\teye = " + gm_to_string( c->cs.eye) + "\n";

	return tmp;
}

gmCamera *GraphicsEngine::get_camera( const std::string &key)
{
	std::map <const std::string, gmCamera *> :: const_iterator iter =
		m_camera.find( key);
	if( iter == m_camera.end()) return 0;
	return (iter->second);
}

void GraphicsEngine::process()
{
	if( !m_enabled) return;

	gmCamera *cam = current_camera();
	if( cam) cam->apply();

	// FIXME: It is possible for the root_renderer "thing" to be delete,
	//	then have another thing take its place before the graphics
	//	can check the "is_free()" function. What will result is that
	//	the new thing will have its processing function called even
	//	though it may not be a graphics processor. The current solution
	//	is to make sure the previous root_renderer notifies the
	//	graphics engine when it is going to be deleted.
	if( root_renderer) {
		if( !(root_renderer->is_free())) {
			root_renderer->m_thing->process(
				game_engine->m_thing_manager);

		} else root_renderer = false;
	}

	// XXX: Some debugging stuff to draw
	if( m_draw_sectors) {
		game_engine->phy_engine->m_sector_map.draw(
			m_draw_sector_level, m_sector_line_width);
	}

	SDL_GL_SwapBuffers();

}

void GraphicsEngine::init_gl_state()
{
	glClearColor( 1, 1, 1, 1);

	glFrontFace( GL_CW);
	glCullFace( GL_BACK);
	glEnable( GL_CULL_FACE);

	glEnable( GL_DEPTH_TEST);
	glClearDepth( 1);
	glDepthFunc( GL_LESS);

	glEnable( GL_BLEND);
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable( GL_TEXTURE_2D);
	//*
	glEnable( GL_LIGHTING);
	rgGLLight li;
	li.set_pos( 0, 1, 0, 0);
	li.set_amb( 0.5, 0.5, 0.5, 1);
	li.set_dif( 1, 1, 1, 1);
	li.set_spec( 1, 1, 1, 1);
	li.enable();
	//*/
}

void GraphicsEngine::set_projection()
{
	glMatrixMode( GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 90.0f, 1.0f, 4.0f, 4096.0f);
	glMatrixMode( GL_MODELVIEW);
}

bool GraphicsEngine::set_root_renderer( const std::string &key)
{
	// FIXME: Notfiy current root_renderer that they are no longer
	// the root renderer?
	ThingControlBlock *tcb = game_engine->m_thing_manager->get( key);
	if( !tcb) return false;
	root_renderer = tcb;
	return true;
}

ModelHandle::ModelHandle( const char *f)
{
	strncpy( fname, f, MAX_MODEL_FILENAME_LENGTH - 1);
	fname[ MAX_MODEL_FILENAME_LENGTH - 1] = '\0';
	model = new sk3DRomModel();

	m_ref_count = 0;
}

ModelHandle::~ModelHandle()
{
	if( reference_count() != 0) {	// FIXME: Should be ASSERT?
		std::cerr << "~ModelHandle ( 0x" << ((void *)this) <<
			") ref_count = " << reference_count() <<
			", model = ( " << ((void *)model) <<
			", \"" << fname << "\")\n";
	}
	if( model) delete model;
}

void ModelHandle::build_bone_name_map()
{
	m_bone.clear();
	if( !model) return;
	typedef std::pair <const std::string, int> SIPair;
	for( int i = 0; i < model->skel.bone_count(); i++) {
		m_bone.insert( SIPair( std::string( model->skel.b[i].bname),
			i));
	}
}

void ModelHandle::build_frame_name_map()
{
	m_frame.clear();
	if( !model) return;
	typedef std::pair <const std::string, int> SIPair;
	for( int i = 0; i < model->skel.frame_count(); i++) {
		m_frame.insert( SIPair( std::string( model->skel.sf[i].fname),
			i));
	}
}

void ModelHandle::build_texture_name_map()
{
	m_texture.clear();
	if( !model) return;
	typedef std::pair <const std::string, int> SIPair;
	for( int i = 0; i < model->texture_count(); i++) {
		m_texture.insert( SIPair( std::string(
			(model->texture[i]).m_tname), i));
	}
}

void ModelHandle::build_group_name_map()
{
	m_group.clear();
	if( !model) return;
	typedef std::pair <const std::string, int> SIPair;
	for( int i = 0; i < model->group_count(); i++) {
		m_group.insert( SIPair( std::string( model->group[i].gname),
			i));
	}
}

void ModelHandle::build_sequence_name_map()
{
	m_sequence.clear();
	if( !model) return;
	typedef std::pair <const std::string, int> SIPair;
	for( int i = 0; i < model->sequence_count(); i++) {
		m_sequence.insert( SIPair( std::string(
			(model->sequence( i))->get_name()), i));
	}
}

int ModelHandle::load()
{
	if( !model) return -1;
	const int err = model->load( fname);
	build_bone_name_map();
	build_frame_name_map();
	build_texture_name_map();
	build_group_name_map();
	build_sequence_name_map();
	if( err) return err;
	return 0;
}

int ModelHandle::set_alternate_texture( const std::string &des,
	const std::string &src, GraphicsEngine *gfx)
{
	if( !model) return -3;
	const int tid = texture_id( des);
	if( tid < 0) return -4;
	
	{ // Unset the current alternate texture if it will be overwritten.
		std::map <int, const std::string> :: iterator iter =
			m_alt_texture.find( tid);
		if( iter != m_alt_texture.end()) {
			TextureHandle *tmp = gfx->get_texture( iter->second);
			if( !tmp) return -5;
			const rgGLTexture *alt_tex = (model->texture[tid]).
				alternate_texture();
			if( alt_tex != tmp->m_tex) {
				std::cerr << "ModelHandle(" << this <<
						")::set_alternate"
					"_texture: model->texture[" << tid
					<< "].alternate_texture() = " <<
					alt_tex << " != " <<
					tmp->m_tex << "\n"
					"Someone (or thing) set the "
						"alternate texture of "
						"ModelHandle(" << this
					<< ")::model = " << model <<
					", which should ONLY be set by " <<
					"ModelHandle(" << this << ")\n"
					"Texture id " << tid << " is no longer"
						" valid\n"
					"Also, ModelHandle(" << this <<
					") will NOT give up its reference to "
					"texture \"" + (iter->second) + "\"\n";
				return -100;
			}
			tmp->dec_reference();
			m_alt_texture.erase( iter);
			(model->texture[tid]).set_alternate_texture( 0);
		}
	}

	// If no texture is being set, then finish.
	TextureHandle *th = gfx->get_texture( src);
	if( !th) return 0;
	if( !(th->m_tex)) return 0;

	typedef std::pair <int, const std::string> SHPair;
	if( false == (m_alt_texture.insert( SHPair( tid, src))).second)
		return -6;

	// Set alternate texture and mark source texture as being referenced.
	(model->texture[tid]).set_alternate_texture( th->m_tex);
	th->inc_reference();
	
	return 0;
}

std::string ModelHandle::dump()
{
	return std::string( "Model: \"" + std::string( fname) +
		"\"\n\tmodel pointer = " + gm_to_string( model) +
		"\n\treference count = " + gm_to_string( m_ref_count) +
		"\n\t" + dump_bones() +
		"\t" + dump_textures() +
		"\t" + dump_groups() +
		"\t" + dump_frames() +
		"\t" + dump_sequences());
}

std::string ModelHandle::dump_bones()
{
	std::string tmp( "Total Bones: " + gm_to_string( int( m_bone.size())) +
		"\n");
	std::map <const std::string, int> :: const_iterator iter =
		m_bone.begin();
	for( ; iter != m_bone.end(); iter++)
		tmp += "\t\t" + gm_to_string( iter->second) + ", \"" +
			(iter->first) + "\"\n";
	return tmp;
}

std::string ModelHandle::dump_textures()
{
	std::string tmp( "Total Textures: " + gm_to_string(
		int( m_texture.size())) + "\n");
	std::map <const std::string, int> :: const_iterator iter =
		m_texture.begin();
	for( ; iter != m_texture.end(); iter++)
		tmp += "\t\t" + gm_to_string( iter->second) + ", \"" +
			(iter->first) + "\"\n";
	return tmp;
}

std::string ModelHandle::dump_groups()
{
	std::string tmp( "Total Groups: " + gm_to_string(
		int( m_group.size())) + "\n");
	std::map <const std::string, int> :: const_iterator iter =
		m_group.begin();
	for( ; iter != m_group.end(); iter++)
		tmp += "\t\t" + gm_to_string( iter->second) + ", \"" +
			(iter->first) + "\"\n";
	return tmp;
}

std::string ModelHandle::dump_frames()
{
	std::string tmp( "Total Frames: " + gm_to_string( int( m_frame.size()))
		+ "\n");
	std::map <const std::string, int> :: const_iterator iter =
		m_frame.begin();
	for( ; iter != m_frame.end(); iter++)
		tmp += "\t\t" + gm_to_string( iter->second) + ", \"" +
			(iter->first) + "\"\n";
	return tmp;
}

std::string ModelHandle::dump_sequences()
{
	std::string tmp( "Total Sequences: " + gm_to_string(
		int( m_sequence.size())) + "\n");
	std::map <const std::string, int> :: const_iterator iter =
		m_sequence.begin();
	for( ; iter != m_sequence.end(); iter++)
		tmp += "\t\t" + gm_to_string( iter->second) + ", \"" +
			(iter->first) + "\"\n";
	return tmp;
}

ModelManager::~ModelManager()
{
	clear();
}

void ModelManager::clear()
{
	while( !(m_model.empty())) {
		// XXX: "m_model" contains key's which ALWAYS point to
		// allocated ModelHandle's. (i.e., "(m_model.begin())->second"
		// will NEVER be null.
		// FIXME: Perhaps add an ASSERT for the above reason?
		delete ((m_model.begin())->second);
		m_model.erase( m_model.begin());
	}
}

ModelHandle *ModelManager::add( const std::string &key,
	const std::string &fname)
{
	ModelHandle *p = new ModelHandle( fname.c_str());
	if( !p) return 0;
	typedef std::pair <const std::string, ModelHandle *> SHPair;
	if( false == (m_model.insert( SHPair( key, p))).second) {
		delete p;
		return 0;
	}
	return p;
}

ModelHandle *ModelManager::get( const std::string &key)
{
	std::map <const std::string, ModelHandle *> :: const_iterator iter =
		m_model.find( key);
	if( iter == m_model.end()) return 0;
	return iter->second;
}

std::map <const std::string, ModelHandle *> :: const_iterator
	ModelManager::get_begin_iterator()
{
	return m_model.begin();
}

std::map <const std::string, ModelHandle *> :: const_iterator
	ModelManager::get_end_iterator()
{
	return m_model.end();
}

int ModelManager::remove( const std::string &key)
{
	std::map <const std::string, ModelHandle *> :: iterator iter =
		m_model.find( key);
	if( iter == m_model.end()) return MODEL_MANAGER_ERR_NO_MODEL;
	ModelHandle *p = iter->second;
	if( p->reference_count() > 0) return MODEL_MANAGER_ERR_REFERENCE;
	delete p;
	m_model.erase( iter);
	return 0;
}

ModelHandle *ModelManager::has_model( const sk3DRomModel *p_model)
{
	std::map <const std::string, ModelHandle *> :: iterator iter =
		m_model.begin();
	for( ; iter != m_model.end(); iter++)
		if( p_model == (iter->second)->model) return (iter->second);
	return 0;
}

TextureHandle::TextureHandle()
{
	m_tex = new rgGLTexture();
	m_ref_count = 0;
}

TextureHandle::~TextureHandle()
{
	if( m_ref_count <= 0) {
		if( m_tex) delete m_tex;

	} else {
		std::cerr << "~Texture(): Reference count = " <<
			m_ref_count << "\n\tCannot delete texture " <<
			m_tex << "\n";
	}
}

TextureHandle *GraphicsEngine::add_texture( const std::string &key)
{
	TextureHandle *t = get_texture( key);
	if( t) return t;
	t = new TextureHandle();
	if( !t) return 0;
	typedef std::pair <const std::string, TextureHandle *> SHPair;
	if( false == (m_texture.insert( SHPair( key, t))).second) {
		delete t;
		return 0;
	}
	return t;
}

/*
int GraphicsEngine::load_texture( const std::string &key,
	const std::string &fname,
	const std::vector < std::string> &args)
{
	std::map <const std::string, TextureHandle *> :: const_iterator iter =
		m_texture.find( key);
	if( iter == m_texture.end()) return -1;
	TextureHandle *t = (iter->second);
	if( !(t->m_tex)) return -2;
	int flags = 0;
	for( int i = 0; i < int( args.size()); i++) {
	}
	const int err = t->m_tex->load( fname.c_str(), flags);
	if( RGERR_SUCCESS != err) return -3;
	return 0;
}
*/

int GraphicsEngine::delete_texture( const std::string &key)
{
	std::map <const std::string, TextureHandle *> :: iterator iter =
		m_texture.find( key);
	return delete_texture( iter);
//	std::map <const std::string, TextureHandle *> :: iterator iter =
//		m_texture.find( key);
//	if( iter == m_texture.end()) return -1;
//	delete (iter->second);
//	m_texture.erase( iter);
//	return 0;
}

int GraphicsEngine::delete_texture(
	std::map <const std::string, TextureHandle *> :: iterator &iter)
{
	if( iter == m_texture.end()) return -1;
	delete (iter->second);
	m_texture.erase( iter);
	return 0;
}

TextureHandle *GraphicsEngine::get_texture( const std::string &key)
{
	std::map <const std::string, TextureHandle *> :: const_iterator iter =
		m_texture.find( key);
	if( iter == m_texture.end()) return 0;
	return (iter->second);
}
